# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Modified by Tack Verification, 2020


"""Module for handling transient state of a running test.

Classes implemented in this module encapsulate state information about a
running test, including test-wide state and currently executing phase
state.

These classes also implement various logic and audit mechanisms for state
transitions during the course of the lifetime of a single Execute()
invocation of an openhtf.Test instance.
"""

import collections
import contextlib
import copy
import functools
import logging
import mimetypes
import os
import socket
import sys
import traceback

from enum import Enum
import mutablerecords

import openhtf
from openhtf import plugs
from openhtf import util
from openhtf.core import measurements
from openhtf.core import phase_executor
from openhtf.core import test_record
from openhtf.util import conf
from openhtf.util import data
from openhtf.util import logs
from past.builtins import long
import six

conf.declare('allow_unset_measurements', default_value=False,
             description='If True, unset measurements do not cause Tests to '
             'FAIL.')
# All tests require a station_id.  This can be via the --config-file
# automatically loaded by OpenHTF, provided explicitly to the config with
# conf.load(station_id='My_OpenHTF_Station'), or alongside other configs loaded
# with conf.load_from_dict({..., 'station_id': 'My_Station'}).  If none of those
# are provided then we'll fall back to the machine's hostname.
conf.declare('station_id', 'The name of this test station',
             default_value=socket.gethostname())

_LOG = logging.getLogger(__name__)

# Sentinel value indicating that the mimetype should be inferred.
INFER_MIMETYPE = object()


class BlankDutIdError(Exception):
  """DUT serial cannot be blank at the end of a test."""


class DuplicateAttachmentError(Exception):
  """Raised when two attachments are attached with the same name."""


class ImmutableMeasurement(collections.namedtuple(
    'ImmutableMeasurement',
    ['name', 'value', 'units', 'dimensions', 'outcome'])):
  """Immutable copy of a measurement."""

  @classmethod
  def FromMeasurement(cls, measurement):
    """Convert a Measurement into an ImmutableMeasurement."""
    measured_value = measurement.measured_value
    if isinstance(measured_value, measurements.DimensionedMeasuredValue):
      value = mutablerecords.CopyRecord(
          measured_value,
          value_dict=copy.deepcopy(measured_value.value_dict)
      )
    else:
      value = (copy.deepcopy(measured_value.value)
               if measured_value.is_value_set else None)

    return cls(
        measurement.name,
        value,
        measurement.units,
        measurement.dimensions,
        measurement.outcome)


class TestState(util.SubscribableStateMixin):
  """This class handles tracking the state of a running Test.

  This class encapsulates all the interesting transient bits of a running Test,
  as opposed to the openhtf.TestDescriptor class, which encapsulates static
  data associated with a Test (that is, it remains the same across invocations
  of Test.Execute()).

  Init Args:
    test_desc: openhtf.TestDescriptor instance describing the test to run,
        used to initialize some values here, but it is not modified.
    execution_uid: a unique uuid use to identify a test being run.
    test_options: test_options passed through from Test.

  Attributes:
    test_record: TestRecord instance for the currently running test.
    state_logger: Logger that logs to test_record's log_records attribute.
    running_phase_state: PhaseState object for the currently running phase,
        if any, otherwise None.
    user_defined_state: Dictionary for users to persist state across phase
        invocations.  It's passed to the user via test_api.
    test_api: An openhtf.TestApi instance for passing to test phases,
        providing test authors access to necessary state information, while
        protecting internal-only structures from being accidentally modified.
        Note that if there is no running phase, test_api is also None.
    execution_uid: A UUID that is specific to this execution.
  """
  Status = Enum('Status', ['WAITING_FOR_TEST_START', 'RUNNING', 'COMPLETED'])  # pylint: disable=invalid-name

  def __init__(self, test_desc, execution_uid, test_options):
    super(TestState, self).__init__()
    self._status = self.Status.WAITING_FOR_TEST_START

    self.test_record = test_record.TestRecord(
        dut_id=None, station_id=conf.station_id, code_info=test_desc.code_info,
        start_time_millis=0,
        # Copy metadata so we don't modify test_desc.
        metadata=copy.deepcopy(test_desc.metadata))
    logs.initialize_record_handler(
        execution_uid, self.test_record, self.notify_update)
    self.state_logger = logs.get_record_logger_for(execution_uid)
    self.plug_manager = plugs.PlugManager(
        test_desc.plug_types, self.state_logger)
    self.running_phase_state = None
    self.user_defined_state = {}
    self.execution_uid = execution_uid
    self.test_options = test_options

  def close(self):
    """Close and remove any global registrations.

    Always call this function when finished with this instance as it ensures
    that it can be garbage collected.

    This function is defined instead of a __del__ function because Python calls
    the __del__ function unreliably.
    """
    logs.remove_record_handler(self.execution_uid)

  @property
  def logger(self):
    if self.running_phase_state:
      return self.running_phase_state.logger
    raise RuntimeError(
        'Calling `logger` attribute while phase not running; use state_logger '
        'instead.')

  @property
  def test_api(self):
    """Create a TestApi for access to this TestState.

    The returned TestApi should be passed as the first argument to test
    phases.  Note that the return value is None if there is no
    self.running_phase_state set.  As such, this attribute should only
    be accessed within a RunningPhaseContext().

    Returns:
      openhtf.TestApi
    """
    running_phase_state = self.running_phase_state
    return (running_phase_state and
            openhtf.TestApi(
                self.logger, self.user_defined_state, self.test_record,
                measurements.Collection(running_phase_state.measurements),
                running_phase_state.attachments,
                running_phase_state.attach,
                running_phase_state.attach_from_file,
                self.get_measurement,
                self.get_attachment,
                self.notify_update,))

  def get_attachment(self, attachment_name):
    """Get a copy of an attachment from current or previous phases.

    Args:
      attachment_name:  str of the attachment name

    Returns:
      A copy of the attachment or None if the attachment cannot be found.

    """
    # Check current running phase state
    if self.running_phase_state:
      if attachment_name in self.running_phase_state.phase_record.attachments:
        attachment = self.running_phase_state.phase_record.attachments.get(
            attachment_name)
        return copy.deepcopy(attachment)

    for phase_record in self.test_record.phases:
      if attachment_name in phase_record.attachments:
        attachment = phase_record.attachments[attachment_name]
        return copy.deepcopy(attachment)

    self.state_logger.warning('Could not find attachment: %s', attachment_name)
    return None

  def get_measurement(self, measurement_name):
    """Get a copy of a measurement value from current or previous phase.

    Measurement and phase name uniqueness is not enforced, so this method will
    return an immutable copy of the most recent measurement recorded.

    Args:
      measurement_name: str of the measurement name
    Returns:
      an ImmutableMeasurement or None if the measurement cannot be found.
    """
    # The framework ignores measurements from SKIP and REPEAT phases
    ignore_outcomes = {test_record.PhaseOutcome.SKIP}

    # Check current running phase state
    if self.running_phase_state:
      if measurement_name in self.running_phase_state.measurements:
        return ImmutableMeasurement.FromMeasurement(
            self.running_phase_state.measurements[measurement_name])

    # Iterate through phases in reversed order to return most recent (necessary
    # because measurement and phase names are not necessarily unique)
    for phase_record in reversed(self.test_record.phases):
      if (phase_record.result not in ignore_outcomes and
          measurement_name in phase_record.measurements):
        measurement = phase_record.measurements[measurement_name]
        return ImmutableMeasurement.FromMeasurement(measurement)

    self.state_logger.warning(
        'Could not find measurement: %s', measurement_name)
    return None

  @contextlib.contextmanager
  def running_phase_context(self, phase_desc):
    """Create a context within which a single phase is running.

    Yields a PhaseState object for tracking transient state during the
    execution of the phase, including the output PhaseRecord.  That PhaseState
    provides the TestApi to be passed into the test phase.

    Within this context, the Station API will report the given phase as the
    currently running phase.

    Args:
      phase_desc: openhtf.PhaseDescriptor to start a context for.

    Yields:
      PhaseState to track transient state.
    """
    assert not self.running_phase_state, 'Phase already running!'
    phase_logger = self.state_logger.getChild('phase.' + phase_desc.name)
    phase_state = self.running_phase_state = PhaseState.from_descriptor(
        phase_desc, self.notify_update, logger=phase_logger)
    try:
      with phase_state.record_timing_context:
        self.notify_update()  # New phase started.
        yield phase_state
    finally:
      self.test_record.add_phase_record(phase_state.phase_record)
      self.running_phase_state = None
      self.notify_update()  # Phase finished.

  def as_base_types(self):
    """Convert to a dict representation composed exclusively of base types."""
    running_phase_state = None
    if self.running_phase_state:
      running_phase_state = self.running_phase_state.as_base_types()
    return {
        'status': data.convert_to_base_types(self._status),
        'test_record': self.test_record.as_base_types(),
        'plugs': self.plug_manager.as_base_types(),
        'running_phase_state': running_phase_state,
    }

  def _asdict(self):
    """Return a dict representation of the test's state."""
    return self.as_base_types()

  @property
  def is_finalized(self):
    return self._status == self.Status.COMPLETED

  def stop_running_phase(self):
    """Stops the currently running phase, allowing another phase to run."""
    self.running_phase_state = None

  @property
  def last_run_phase_name(self):
    """Get the name of the currently running phase, or None.

    Note that this name is not guaranteed to still be accurate by the time this
    method returns, so this should only be used for log messages/user display
    and not for programmatic purposes.

    Returns:
      str name of currently running phase or None.
    """
    return self.running_phase_state and self.running_phase_state.name

  def mark_test_started(self):
    """Set the TestRecord's start_time_millis field."""
    # Blow up instead of blowing away a previously set start_time_millis.
    assert self.test_record.start_time_millis is 0
    self.test_record.start_time_millis = util.time_millis()
    self.notify_update()

  def set_status_running(self):
    """Mark the test as actually running, can't be done once finalized."""
    if self._is_aborted():
      return
    assert self._status == self.Status.WAITING_FOR_TEST_START
    self._status = self.Status.RUNNING
    self.notify_update()

  def finalize_from_phase_outcome(self, phase_execution_outcome):
    """Finalize due to the given phase outcome.

    Args:
      phase_execution_outcome: An instance of
          phase_executor.PhaseExecutionOutcome.
    """
    if self._is_aborted():
      return

    # Handle a few cases where the test is ending prematurely.
    if phase_execution_outcome.raised_exception:
      result = phase_execution_outcome.phase_result
      if isinstance(result, phase_executor.ExceptionInfo):
        code = result.exc_type.__name__
        description = str(result.exc_val)
      else:
        # openhtf.util.threads.ThreadTerminationError gets str'd directly.
        code = str(type(phase_execution_outcome.phase_result).__name__)
        description = str(phase_execution_outcome.phase_result)
      self.test_record.add_outcome_details(code, description)
      if self._outcome_is_failure_exception(phase_execution_outcome):
        self.state_logger.error(
            'Outcome will be FAIL since exception was of type %s',
            phase_execution_outcome.phase_result.exc_type)
        self._finalize(test_record.Outcome.FAIL)
      else:
        self.state_logger.critical(
            'Finishing test execution early due to an exception raised during '
            'phase execution; outcome ERROR.')
        # Enable CLI printing of the full traceback with the -v flag.
        self.state_logger.critical(
            'Traceback:%s%s%s%s',
            os.linesep,
            ''.join(traceback.format_tb(
                phase_execution_outcome.phase_result.exc_tb)),
            os.linesep,
            description,
        )
        self._finalize(test_record.Outcome.ERROR)
    elif phase_execution_outcome.is_timeout:
      self.state_logger.error('Finishing test execution early due to '
                              'phase timeout, outcome TIMEOUT.')
      self.test_record.add_outcome_details('TIMEOUT',
                                           'A phase hit its timeout.')
      self._finalize(test_record.Outcome.TIMEOUT)
    elif phase_execution_outcome.phase_result == openhtf.PhaseResult.STOP:
      self.state_logger.error('Finishing test execution early due to '
                              'PhaseResult.STOP, outcome FAIL.')
      self.test_record.add_outcome_details('STOP',
                                           'A phase stopped the test run.')
      self._finalize(test_record.Outcome.FAIL)

  def finalize_normally(self):
    """Mark the state as finished.

    This method is called on normal test completion. The outcome will be either
    PASS, FAIL, or ERROR, depending on phases' outcomes.
    """
    if self._is_aborted():
      return

    phases = self.test_record.phases
    if not phases:
      # Vacuously PASS a TestRecord with no phases.
      self._finalize(test_record.Outcome.PASS)
    elif any(
        phase.outcome == test_record.PhaseOutcome.FAIL for phase in phases):
      # Any FAIL phase results in a test failure.
      self._finalize(test_record.Outcome.FAIL)
    elif all(
        phase.outcome == test_record.PhaseOutcome.SKIP for phase in phases):
      # Error when all phases are skipped; otherwise, it could lead to
      # unintentional passes.
      self.state_logger.error('All phases were skipped, outcome ERROR.')
      self.test_record.add_outcome_details(
          'ALL_SKIPPED', 'All phases were unexpectedly skipped.')
      self._finalize(test_record.Outcome.ERROR)
    else:
      # Otherwise, the test run was successful.
      self._finalize(test_record.Outcome.PASS)

    self.state_logger.debug(
        'Finishing test execution normally with outcome %s.',
        self.test_record.outcome.name)

  def abort(self):
    if self._is_aborted():
      return

    self.state_logger.debug('Finishing test execution early due to '
                            'test abortion, outcome ABORTED.')
    self.test_record.add_outcome_details('ABORTED', 'Test aborted by operator.')
    self._finalize(test_record.Outcome.ABORTED)

  def _finalize(self, test_outcome):
    aborting = test_outcome == test_record.Outcome.ABORTED
    assert not self.is_finalized or aborting, (
        'Test already completed with status %s!' % self._status.name)

    self.test_record.outcome = test_outcome
    
    if self.test_record.is_started():
      self.test_record.end_time_millis = util.time_millis()
    else:
      self.test_record.end_time_millis = self.test_record.start_time_millis
      
    self._status = self.Status.COMPLETED
    self.notify_update()

  def _is_aborted(self):
    if (self.is_finalized and
        self.test_record.outcome == test_record.Outcome.ABORTED):
      self.state_logger.debug('Test already aborted.')
      return True
    return False

  def _outcome_is_failure_exception(self, outcome):
    for failure_exception in self.test_options.failure_exceptions:
      if isinstance(outcome.phase_result.exc_val, failure_exception):
        return True
    return False

  def __str__(self):
    return '<%s: %s@%s Running Phase: %s>' % (
        type(self).__name__,
        self.test_record.dut_id,
        self.test_record.station_id,
        self.last_run_phase_name,
    )


class PhaseState(
    mutablerecords.Record('PhaseState',
                          ['name', 'phase_record', 'measurements', 'options'],
                          {'hit_repeat_limit': False,
                           'notify_cb': None,
                           'logger': None,
                           '_cached': dict,
                           '_update_measurements': set})):
  """Data type encapsulating interesting information about a running phase.

  Attributes:
    phase_record: A test_record.PhaseRecord for the running phase.
    attachments: Convenience accessor for phase_record.attachments.
    measurements: A dict mapping measurement name to it's declaration; this
        dict can be passed to measurements.Collection to initialize a user-
        facing Collection for setting measurements.
    options: the PhaseOptions from the phase descriptor.
    result: Convenience getter/setter for phase_record.result.
    _cached: A cached representation of the running test state that; updated in
        place to save allocation time.
  """

  def __init__(self, *args, **kwargs):
    super(PhaseState, self).__init__(*args, **kwargs)
    for m in six.itervalues(self.measurements):
      # Using functools.partial to capture the value of the loop variable.
      m.set_notification_callback(functools.partial(self._notify, m.name))
    self._cached = {
        'name': self.name,
        'codeinfo': data.convert_to_base_types(self.phase_record.codeinfo),
        'descriptor_id': data.convert_to_base_types(
            self.phase_record.descriptor_id),
        # Options are not set until the phase is finished.
        'options': None,
        'measurements': {
            k: m.as_base_types() for k, m in six.iteritems(self.measurements)},
        'attachments': {},
    }

  @classmethod
  def from_descriptor(cls, phase_desc, notify_cb, logger=None):
    return cls(
        phase_desc.name,
        test_record.PhaseRecord.from_descriptor(phase_desc),
        collections.OrderedDict(
            (measurement.name, copy.deepcopy(measurement))
            for measurement in phase_desc.measurements),
        phase_desc.options,
        notify_cb=notify_cb,
        logger=logger)

  def _notify(self, measurement_name):
    self._update_measurements.add(measurement_name)
    if self.notify_cb:
      self.notify_cb()

  def as_base_types(self):
    """Convert to a dict representation composed exclusively of base types."""
    cur_update_measurements = self._update_measurements
    self._update_measurements = set()
    # Update the dictionaries previously returned for the measurements that
    # have been updated.
    for m in cur_update_measurements:
      self.measurements[m].as_base_types()
    return self._cached

  @property
  def result(self):
    return self.phase_record.result

  @result.setter
  def result(self, result):
    self.phase_record.result = result

  @property
  def attachments(self):
    return self.phase_record.attachments

  def attach(self, name, binary_data, mimetype=INFER_MIMETYPE):
    """Store the given binary_data as an attachment with the given name.

    Args:
      name: Attachment name under which to store this binary_data.
      binary_data: Data to attach.
      mimetype: One of the following:
          INFER_MIMETYPE - The type will be guessed from the attachment name.
          None - The type will be left unspecified.
          A string - The type will be set to the specified value.

    Raises:
      DuplicateAttachmentError: Raised if there is already an attachment with
        the given name.
    """
    if name in self.phase_record.attachments:
      raise DuplicateAttachmentError('Duplicate attachment for %s' % name)

    if mimetype is INFER_MIMETYPE:
      mimetype = mimetypes.guess_type(name)[0]
    elif mimetype is not None and not mimetypes.guess_extension(mimetype):
      _LOG.warning('Unrecognized MIME type: "%s" for attachment "%s"',
                   mimetype, name)

    attach_record = test_record.Attachment(binary_data, mimetype)
    self.phase_record.attachments[name] = attach_record
    self._cached['attachments'][name] = attach_record._asdict()

  def attach_from_file(self, filename, name=None, mimetype=INFER_MIMETYPE):
    """Store the contents of the given filename as an attachment.

    Args:
      filename: The file to read data from to attach.
      name: If provided, override the attachment name, otherwise it will
        default to the filename.
      mimetype: One of the following:
          INFER_MIMETYPE: The type will be guessed first, from the file name,
              and second (i.e. as a fallback), from the attachment name.
          None: The type will be left unspecified.
          A string: The type will be set to the specified value.

    Raises:
      DuplicateAttachmentError: Raised if there is already an attachment with
        the given name.
      IOError: Raised if the given filename couldn't be opened.
    """
    if mimetype is INFER_MIMETYPE:
      mimetype = mimetypes.guess_type(filename)[0] or mimetype
    with open(filename, 'rb') as f:  # pylint: disable=invalid-name
      self.attach(
          name if name is not None else os.path.basename(filename), f.read(),
          mimetype=mimetype)

  def _finalize_measurements(self):
    """Perform end-of-phase finalization steps for measurements.

    Any UNSET measurements will cause the Phase to FAIL unless
    conf.allow_unset_measurements is set True.
    """
    for measurement in self.measurements.values():
      # Clear notification callbacks for later serialization.
      measurement.set_notification_callback(None)
      # Validate multi-dimensional measurements now that we have all values.
      if measurement.outcome is measurements.Outcome.PARTIALLY_SET:
        try:
          measurement.validate()
        except Exception:  # pylint: disable=broad-except
          # Record the exception as the new result.
          if self.phase_record.result.is_terminal:
            _LOG.exception(
                'Measurement validation raised an exception, but phase result '
                'is already terminal; logging additional exception here.')
          else:
            self.phase_record.result = phase_executor.PhaseExecutionOutcome(
                phase_executor.ExceptionInfo(*sys.exc_info()))

    # Set final values on the PhaseRecord.
    self.phase_record.measurements = self.measurements

  def _measurements_pass(self):
    allowed_outcomes = {measurements.Outcome.PASS}
    if conf.allow_unset_measurements:
      allowed_outcomes.add(measurements.Outcome.UNSET)

    if any(meas.outcome not in allowed_outcomes
           for meas in self.phase_record.measurements.values()):
      return False
    return True

  def _set_phase_outcome(self):
    if self.result is None or self.result.is_terminal or self.hit_repeat_limit:
      outcome = test_record.PhaseOutcome.ERROR
    elif self.result.is_repeat or self.result.is_skip:
      outcome = test_record.PhaseOutcome.SKIP
    elif self.result.is_fail_and_continue:
      outcome = test_record.PhaseOutcome.FAIL
    else:
      outcome = (test_record.PhaseOutcome.PASS
                 if self._measurements_pass()
                 else test_record.PhaseOutcome.FAIL)
    self.phase_record.outcome = outcome

  @property
  @contextlib.contextmanager
  def record_timing_context(self):
    """Context manager for the execution of a single phase.

    This method performs some pre-phase setup on self (for measurements), and
    records the start and end time based on when the context is entered/exited.

    Yields:
      None
    """
    self._cached['start_time_millis'] = long(
        self.phase_record.record_start_time())

    try:
      yield
    finally:
      self._finalize_measurements()
      self._set_phase_outcome()
      self.phase_record.finalize_phase(self.options)



"""Serves an Angular frontend and information about a running OpenHTF test.

This server does not currently support more than one test running in the same
process. However, the dashboard server (dashboard_server.py) can be used to
aggregate info from multiple station servers with a single frontend.
"""

import contextlib
import itertools
import json
import logging
import os
import re
import six
import socket
import threading
import time
import types
import collections

import sockjs.tornado

import openhtf
from openhtf.output.callbacks import mfg_inspector
from openhtf.output.servers import pub_sub
from openhtf.output.servers import web_gui_server
from openhtf.util import conf
from openhtf.util import data
from openhtf.util import functions
from openhtf.util import logs
from openhtf.util import multicast
from openhtf.util import timeouts

STATION_SERVER_TYPE = 'station'

MULTICAST_QUERY = 'OPENHTF_DISCOVERY'
TEST_STATUS_COMPLETED = 'COMPLETED'

_LOG = logging.getLogger(__name__)

# Constants related to response times within the server.
_CHECK_FOR_FINISHED_TEST_POLL_S = 0.5
_DEFAULT_FRONTEND_THROTTLE_S = 0.15
_WAIT_FOR_ANY_EVENT_POLL_S = 0.05
_WAIT_FOR_EXECUTING_TEST_POLL_S = 0.1

conf.declare('frontend_throttle_s', default_value=_DEFAULT_FRONTEND_THROTTLE_S,
             description=('Min wait time between successive updates to the '
                          'frontend.'))
conf.declare('station_server_port', default_value=0,
             description=('Port on which to serve the app. If set to zero (the '
                          'default) then an arbitrary port will be chosen.'))

# These have default values in openhtf.util.multicast.py.
conf.declare('station_discovery_address')
conf.declare('station_discovery_port')
conf.declare('station_discovery_ttl')


def _get_executing_test():
  """Get the currently executing test and its state.

  When this function returns, it is not guaranteed that the returned test is
  still running. A consumer of this function that wants to access test.state is
  exposed to a race condition in which test.state may become None at any time
  due to the test finishing. To address this, in addition to returning the test
  itself, this function returns the last known test state.

  Returns:
    test: The test that was executing when this function was called, or None.
    test_state: The state of the executing test, or None.
  """
  tests = list(six.itervalues(openhtf.Test.TEST_INSTANCES))

  if not tests:
    return None, None

  if len(tests) > 1:
    _LOG.warn('Station server does not support multiple executing tests.')

  test = tests[0]
  test_state = test.state

  if test_state is None:
    # This is the case if:
    # 1. The test executor was created but has not started running.
    # 2. The test finished while this function was running, after we got the
    #        list of tests but before we accessed the test state.
    return None, None

  return test, test_state


def _test_state_from_record(test_record_dict, execution_uid=None):
  """Convert a test record dict to a test state dict.

  Args:
    test_record_dict: An OpenHTF TestRecord, converted to base types.
    execution_uid: Execution ID of the running test.

  Returns:
    Dictionary representation of a test's final state. On top of the fields from
        TestState._asdict() we add 'execution_uid' which is needed by the
        frontend app.
  """
  return {
      'execution_uid': execution_uid,
      'plugs': {
          'plug_states': {},
      },
      'running_phase_state': None,
      'status': TEST_STATUS_COMPLETED,
      'test_record': test_record_dict,
  }


def _wait_for_any_event(events, timeout_s):
  """Wait for any in a list of threading.Event's to be set.

  Args:
    events: List of threading.Event's.
    timeout_s: Max duration in seconds to wait before returning.

  Returns:
      True if at least one event was set before the timeout expired, else False.
  """
  def any_event_set():
    return any(event.is_set() for event in events)

  result = timeouts.loop_until_timeout_or_true(
      timeout_s, any_event_set, sleep_s=_WAIT_FOR_ANY_EVENT_POLL_S)

  return result or any_event_set()


class StationWatcher(threading.Thread):
  """Watches for changes in the state of the currently running OpenHTF test.

  The StationWatcher uses an event-based mechanism to detect changes in test
  state. This means we rely on the OpenHTF framework to call notify_update()
  when a change occurs. Authors of frontend-aware plugs must ensure that
  notify_update() is called when a change occurs to that plug's state.
  """
  daemon = True

  def __init__(self, update_callback):
    super(StationWatcher, self).__init__(name=type(self).__name__)
    self._update_callback = update_callback

  def run(self):
    """Call self._poll_for_update() in a loop and handle errors."""
    while True:
      try:
        self._poll_for_update()
      except RuntimeError as error:
        # Note that because logging triggers a call to notify_update(), by
        # logging a message, we automatically retry publishing the update
        # after an error occurs.
        if error.message == 'dictionary changed size during iteration':
          # These errors occur occasionally and it is infeasible to get rid of
          # them entirely unless data.convert_to_base_types() is made
          # thread-safe. Ignore the error and retry quickly.
          _LOG.debug('Ignoring (probably harmless) error in station watcher: '
                     '`dictionary changed size during iteration`.')
          time.sleep(0.1)
        else:
          _LOG.exception('Error in station watcher: %s', error)
          time.sleep(1)
      except Exception as error:  # pylint: disable=broad-except
        _LOG.exception('Error in station watcher: %s', error)
        time.sleep(1)

  @functions.call_at_most_every(float(conf.frontend_throttle_s))
  def _poll_for_update(self):
    """Call the callback with the current test state, then wait for a change."""
    test, test_state = _get_executing_test()

    if test is None:
      time.sleep(_WAIT_FOR_EXECUTING_TEST_POLL_S)
      return

    state_dict, event = self._to_dict_with_event(test_state)
    self._update_callback(state_dict)

    plug_manager = test_state.plug_manager
    plug_events = [
        plug_manager.get_plug_by_class_path(plug_name).asdict_with_event()[1]
        for plug_name in plug_manager.get_frontend_aware_plug_names()
    ]
    events = [event] + plug_events

    # Wait for the test state or a plug state to change, or for the previously
    # executing test to finish.
    while not _wait_for_any_event(events, _CHECK_FOR_FINISHED_TEST_POLL_S):
      new_test, new_test_state = _get_executing_test()
      if test != new_test:
        break
      if test_state.execution_uid != new_test_state.execution_uid:
        # TODO
        
        # This fixes a race condition that sometimes happen between tests
        # causing the last websocket message to contain the last test's uid
        # and forcing a 404 error when the frontend tries to fetch the test phases.

        # For some reason, between new_test, new_test_state, test and new_test,
        # only new_test_state has the old UID which caused the previous if clause
        # to not notice the change of test and enter an infinte loop.

        # This may point to an underlying bug, but this was introduced as a quick fix.
        break

  @classmethod
  def _to_dict_with_event(cls, test_state):
    """Process a test state into the format we want to send to the frontend."""
    original_dict, event = test_state.asdict_with_event()

    # This line may produce a 'dictionary changed size during iteration' error.
    test_state_dict = data.convert_to_base_types(original_dict)

    test_state_dict['execution_uid'] = test_state.execution_uid
    return test_state_dict, event


class DashboardPubSub(sockjs.tornado.SockJSConnection):
  """WebSocket endpoint for the list of available stations.

  In this case, there is always exactly one available station: the station
  running the StationServer. See dashboard_server.py for an implementation of
  the dashboard WebSocket endpoint for multiple stations.

  TODO(Kenadia): Remove this endpoint from the station server. Since the
  frontend knows whether it is running off of a station server or dashboard
  server, it should be smart enough not to look for this endpoint on the station
  server.
  """
  port = None  # Set by for_port().

  @classmethod
  def for_port(cls, port):
    """Returns a new subclass with the port set."""
    return type(cls.__name__, (cls,), {'port': port})

  def on_open(self, unused_info):
    """Called by the base class when a client connects."""
    self.send(self._make_message())

  @classmethod
  def _make_message(cls):
    host = 'localhost'
    host_port = '%s:%s' % (host, cls.port)
    return {
        host_port: {
            'station_id': conf.station_id,  # From openhtf.core.test_state.
            'host': host,
            'port': cls.port,
            'status': 'ONLINE',
        }
    }


class StationPubSub(pub_sub.PubSub):
  """WebSocket endpoint for test updates.

  The endpoint provides information about the test that is currently running
  with this StationServer. Two types of message are sent: 'update' and 'record',
  where 'record' indicates the final state of a test.
  """
  _lock = threading.Lock()  # Required by pub_sub.PubSub.
  subscribers = set()  # Required by pub_sub.PubSub.
  _last_execution_uid = None
  _last_message = None

  @classmethod
  def publish_test_record(cls, test_record):
    test_record_dict = data.convert_to_base_types(test_record)
    test_state_dict = _test_state_from_record(test_record_dict,
                                              cls._last_execution_uid)
    cls._publish_test_state(test_state_dict, 'record')

  @classmethod
  def publish_update(cls, test_state_dict):
    """Publish the state of the currently executing test."""
    cls._publish_test_state(test_state_dict, 'update')

  @classmethod
  def _publish_test_state(cls, test_state_dict, message_type):
    message = {
        'state': test_state_dict,
        'test_uid': test_state_dict['execution_uid'],
        'type': message_type,
    }
    super(StationPubSub, cls).publish(message)
    cls._last_execution_uid = test_state_dict['execution_uid']
    cls._last_message = message

  def on_subscribe(self, info):
    """Send the more recent test state to new subscribers when they connect."""
    if self._last_message is not None:
      self.send(self._last_message)


class BaseTestHandler(web_gui_server.CorsRequestHandler):
  """Base class for HTTP endpoints that get test data."""

  def get_test(self, test_uid):
    """Get the specified test. Write 404 and return None if it is not found."""
    test, test_state = _get_executing_test()

    if test is None or str(test.uid) != test_uid:
      self.write('Unknown test UID %s' % test_uid)
      self.set_status(404)
      return None, None

    return test, test_state


class AttachmentsHandler(BaseTestHandler):
  """GET endpoint for a file attached to a test."""

  def get(self, test_uid, phase_descriptor_id, attachment_name):
    _, test_state = self.get_test(test_uid)

    if test_state is None:
      return

    # Find the phase matching `phase_descriptor_id`.
    running_phase = test_state.running_phase_state
    phase_records = itertools.chain(
        test_state.test_record.phases,
        [running_phase.phase_record] if running_phase is not None else [])

    matched_phase = None
    for phase in phase_records:
      if str(phase.descriptor_id) == phase_descriptor_id:
        matched_phase = phase
        break

    if matched_phase is None:
      self.write('Unknown phase descriptor %s' % phase_descriptor_id)
      self.set_status(404)
      return

    # Find the attachment matching `attachment_name`.
    if attachment_name in matched_phase.attachments:
      attachment = matched_phase.attachments[attachment_name]
    else:
      self.write('Unknown attachment %s' % attachment_name)
      self.set_status(404)
      return

    self.set_header('Content-Type', attachment.mimetype)
    self.write(attachment.data)


class PhasesHandler(BaseTestHandler):
  """GET endpoint for phase descriptors for a test, i.e. the full phase list."""

  def get(self, test_uid):
    test, _ = self.get_test(test_uid)

    if test is None:
      return

    transform_phase_to_dict = lambda phase: dict(id=id(phase), **data.convert_to_base_types(phase))
    
    phase_descriptors = [
        transform_phase_to_dict(phase)
        for phase in test.descriptor.phase_group]
    
    
    
    phase_tree = test.descriptor.phase_group.build_tree()
    def tree_to_dict(tree):
      if isinstance(tree, collections.Iterable):
        return [tree_to_dict(item) for item in tree]
      else:
        return transform_phase_to_dict(tree)
    
    phase_tree_dict = tree_to_dict(phase_tree)
    # Wrap value in a dict because writing a list directly is prohibited.
    self.write({'data': phase_descriptors, 'tree': phase_tree_dict})


class PlugsHandler(BaseTestHandler):
  """POST endpoints to receive plug responses from the frontend."""

  def post(self, test_uid, plug_name):
    _, test_state = self.get_test(test_uid)

    if test_state is None:
      return

    # Find the plug matching `plug_name`.
    plug = test_state.plug_manager.get_plug_by_class_path(plug_name)
    if plug is None:
      self.write('Unknown plug %s' % plug_name)
      self.set_status(404)
      return

    try:
      request = json.loads(self.request.body.decode('utf-8'))
      method_name = request['method']
      args = request['args']
    except (KeyError, ValueError):
      self.write('Malformed JSON request.')
      self.set_status(400)
      return

    method = getattr(plug, method_name, None)

    if not (plug.enable_remote and
            isinstance(method, types.MethodType) and
            not method_name.startswith('_') and
            method_name not in plug.disable_remote_attrs):
      self.write('Cannot access method %s of plug %s.' % (method_name,
                                                          plug_name))
      self.set_status(400)
      return

    try:
      response = json.dumps(method(*args))
    except Exception as e:  # pylint: disable=broad-except
      self.write('Plug error: %s' % repr(e))
      self.set_status(500)
    else:
      self.write(response)


class BaseHistoryHandler(web_gui_server.CorsRequestHandler):

  def initialize(self, history_path):
    self.history_path = history_path


class HistoryListHandler(BaseHistoryHandler):
  """GET endpoint for the list of tests in the history.

  When requesting the history list, we respond with all files in the history
  folder ending with the '.pb' extension. Ideally, file names should match the
  following form (see chtf.py):

      'mfg_event_{dut_id}_{start_time_millis}.pb'

  The requester can filter the returned history items by passing DUT ID and/or
  start time as query parameters.
  """

  def get(self):
    filter_dut_id = self.get_arguments('dutId')
    filter_start_time_millis = self.get_arguments('startTimeMillis')

    history_items = []

    for file_name in os.listdir(self.history_path):
      if not file_name.endswith('.pb'):
        continue

      if not os.path.isfile(os.path.join(self.history_path, file_name)):
        continue

      dut_id = None
      start_time_millis = None
      match = re.match(r'mfg_event_(.+)_(\d+)\.pb$', file_name)

      if match is not None:
        dut_id = match.group(1)
        start_time_millis = int(match.group(2))

      if filter_dut_id and dut_id not in filter_dut_id:
        continue

      if (filter_start_time_millis and
          str(start_time_millis) not in filter_start_time_millis):
        continue

      history_items.append({
          'dut_id': dut_id,
          'file_name': file_name,
          'start_time_millis': start_time_millis,
      })

    # Wrap value in a dict because writing a list directly is prohibited.
    self.write({'data': history_items})


class HistoryItemHandler(BaseHistoryHandler):
  """GET endpoint for a test record from the history."""

  def get(self, file_name):
    # TODO(Kenadia): Implement the history item handler. The implementation
    # depends on the format used to store test records on disk.
    self.write('Not implemented.')
    self.set_status(500)


class HistoryAttachmentsHandler(BaseHistoryHandler):
  """GET endpoint for an attachment from an MfgEvent in the history.

  The sha1 query parameter is optional and used as a backup to identify an
  attachment if the name does not match any known name. Including this parameter
  is recommended, as some systems may modify attachment names when storing them
  on the MfgEvent in the case where multiple attachments have the same name.
  """

  def get(self, file_name, attachment_name):
    # TODO(Kenadia): Implement the history item handler. The implementation
    # depends on the format used to store test records on disk.
    self.write('Not implemented.')
    self.set_status(500)


class StationMulticast(multicast.MulticastListener):
  """Announce the existence of a station server to any searching dashboards."""

  def __init__(self, station_server_port):
    # These have default values in openhtf.util.multicast.py.
    kwargs = {
        attr: conf['station_discovery_%s' % attr]
        for attr in ('address', 'port', 'ttl')
        if 'station_discovery_%s' % attr in conf
    }
    super(StationMulticast, self).__init__(self._make_message, **kwargs)
    self.station_server_port = station_server_port

  def _make_message(self, message):
    if message != MULTICAST_QUERY:
      if message == 'OPENHTF_PING':
        # Don't log for the old multicast string.
        return
      _LOG.debug('Got unexpected traffic on multicast socket: %s', message)
      return

    _, test_state = _get_executing_test()

    if test_state:
      cell = test_state.test_record.metadata.get('cell')
      test_description = test_state.test_record.metadata.get('test_description')
      test_name = test_state.test_record.metadata.get('test_name')
    else:
      cell = None
      test_description = None
      test_name = None

    return json.dumps({
        'cell': cell,
        'port': self.station_server_port,
        'station_id': conf.station_id,  # From openhtf.core.test_state.
        'test_description': test_description,
        'test_name': test_name,
    })


class StationServer(web_gui_server.WebGuiServer):
  """Provides endpoints for interacting with an OpenHTF test.

  Also serves an Angular frontend that interfaces with those endpoints.

  Can be used as a context manager to ensure the server is stopped cleanly:

    with StationServer(history_path) as server:
      test = openhtf.Test(*my_phases)
      test.add_output_callbacks(server.publish_final_state)
      test.execute()

  Can also be used via the maybe_run() helper function:

    with maybe_run(should_run, history_path) as server:
      test = openhtf.Test(*my_phases)
      if server:
        test.add_output_callbacks(server.publish_final_state)
      test.execute()
  """

  def __init__(self, history_path=None, static_files_root=None):
    # Disable tornado's logging.
    # TODO(Kenadia): Enable these logs if verbosity flag is at least -vvv.
    #     I think this will require changing how StoreRepsInModule works.
    #     Currently, if we call logs.ARG_PARSER.parse_known_args() multiple
    #     times, we multiply the number of v's that we get.
    tornado_logger = logging.getLogger('tornado')
    tornado_logger.propagate = False
    tornado_logger.setLevel(logging.INFO)
    
    if not tornado_logger.handlers:
      tornado_logger.addHandler(logging.NullHandler())

    # Bind port early so that the correct port number can be used in the routes.
    sockets, port = web_gui_server.bind_port(int(conf.station_server_port))

    # Set up the station watcher.
    station_watcher = StationWatcher(StationPubSub.publish_update)
    station_watcher.start()

    # Set up the SockJS endpoints.
    dashboard_class = DashboardPubSub.for_port(port)
    dash_router = sockjs.tornado.SockJSRouter(dashboard_class, '/sub/dashboard')
    station_router = sockjs.tornado.SockJSRouter(StationPubSub, '/sub/station')
    routes = dash_router.urls + station_router.urls

    # Set up the other endpoints.
    routes.extend((
        (r'/tests/(?P<test_uid>[\w\d:]+)/phases', PhasesHandler),
        (r'/tests/(?P<test_uid>[\w\d:]+)/plugs/(?P<plug_name>.+)',
         PlugsHandler),
        (r'/tests/(?P<test_uid>[\w\d:]+)/phases/(?P<phase_descriptor_id>\d+)/'
         'attachments/(?P<attachment_name>.+)', AttachmentsHandler),
    ))

    # Optionally enable history from disk.
    if history_path is not None:
      routes.extend((
          (r'/history', HistoryListHandler, {'history_path': history_path}),
          (r'/history/(?P<file_name>[^/]+)', HistoryItemHandler,
           {'history_path': history_path}),
          (r'/history/(?P<file_name>[^/]+)/attachments/(?P<attachment_name>.+)',
           HistoryAttachmentsHandler, {'history_path': history_path}),
      ))

    super(StationServer, self).__init__(routes, port, sockets=sockets, static_files_root=static_files_root)
    self.station_multicast = StationMulticast(port)

  def _get_config(self):
    return {
        'server_type': STATION_SERVER_TYPE,
    }

  def run(self):
    _LOG.info('Announcing station server via multicast on %s:%s',
              self.station_multicast.address, self.station_multicast.port)
    self.station_multicast.start()
    _LOG.info(
        'Starting station server at:\n'
        '  Local: http://localhost:{port}\n'
        '  Remote: http://{host}:{port}'
        .format(host=socket.gethostname(), port=self.port))
    super(StationServer, self).run()

  def stop(self):
    _LOG.info('Stopping station server.')
    super(StationServer, self).stop()
    _LOG.info('Stopping multicast.')
    self.station_multicast.stop(timeout_s=0)

  def publish_final_state(self, test_record):
    """Test output callback publishing a final state from the test record."""
    StationPubSub.publish_test_record(test_record)


@contextlib.contextmanager
def maybe_run(should_run, history_path=None):
  """Provides a context which conditionally runs a StationServer."""
  if not should_run:
    yield
    return
  with StationServer(history_path) as server:
    yield server



import os
import pytest
import tempfile
import shutil

from mock import MagicMock
from spintop_openhtf.plugs.visa import VISAInterface

class InstrumentMock(object):
    timeout = 5000

    def query(self, cmd):
        pass

    def read(self):
        pass

    def write(self, cmd):
        pass

    def write_binary_values(self, cmd, values, datatype=None):
        pass

    def query_binary_values(self, cmd, datatype=None):
        pass

@pytest.fixture()
def instrument_mock():
    return MagicMock(spec=InstrumentMock)

@pytest.fixture()
def interface(instrument_mock):
    interface = VISAInterface()
    interface.open(_instrument=instrument_mock)
    return interface

@pytest.fixture()
def temp_filename():
    TEMP_FOLDER = tempfile.mkdtemp()
    try:
        yield os.path.join(TEMP_FOLDER, 'destination.txt')
    finally:
        shutil.rmtree(TEMP_FOLDER)

def test_execute_command(interface, instrument_mock):
    READ_VALUE = 'Hello'
    instrument_mock.read.return_value = READ_VALUE
    
    read_value = interface.execute_command('write', doread=True)

    assert read_value == READ_VALUE
    instrument_mock.write.assert_called_with('write')

def test_wait_for_command_end(interface, instrument_mock):
    """Should query '*OPC?' when execute_command called without doread."""
    
    interface.execute_command('write')
    instrument_mock.write.assert_called_with('write')
    instrument_mock.query.assert_called_with('*OPC?')

def test_timeout_is_set_ms(interface, instrument_mock):
    TIMEOUT_S = 29
    interface.execute_command('write', timeout=TIMEOUT_S)

    assert instrument_mock.timeout == TIMEOUT_S*1000

def test_read_file(interface, instrument_mock, temp_filename):
    CONTENT = b'something'
    instrument_mock.query_binary_values.return_value = CONTENT

    interface.read_binary_file('command', temp_filename)

    instrument_mock.query_binary_values.assert_called_with('command', datatype='c')

    with open(temp_filename, 'rb') as written_file:
        assert written_file.read() == CONTENT

def test_write_file(interface, instrument_mock, temp_filename):
    CONTENT = b'something'

    with open(temp_filename, 'wb') as file_to_write:
        file_to_write.write(CONTENT)
    
    interface.write_binary_file('command', temp_filename)

    instrument_mock.write_binary_values.assert_called_with('command', list(CONTENT), datatype='c')

def test_reset(interface, instrument_mock):
    """Should send '*RST' and then query '*OPC?'"""

    interface.reset()

    instrument_mock.write.assert_called_with('*RST')
    instrument_mock.query.assert_called_with('*OPC?')




# Copyright 2017 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""OpenHTF base plugs for thinly wrapping existing device abstractions.

Sometimes you already have a Python interface to a device or instrument; you
just need to put that interface in plug form to get it into your test phase.
Device-wrapping plugs are your friends in such times.
"""

import functools

import openhtf
import six


def short_repr(obj, max_len=40):
  """Returns a short, term-friendly string representation of the object.

  Args:
    obj: An object for which to return a string representation.
    max_len: Maximum length of the returned string. Longer reprs will be turned
        into a brief descriptive string giving the type and length of obj.
  """
  obj_repr = repr(obj)
  if len(obj_repr) <= max_len:
    return obj_repr
  return '<{} of length {}>'.format(type(obj).__name__, len(obj_repr))


class DeviceWrappingPlug(openhtf.plugs.BasePlug):
  """A base plug for wrapping existing device abstractions.

  Attribute access is delegated to the _device attribute, which is normally set
  by passing some device instance to the constructor of this base class.
  Subclasses can use the @conf.inject_positional_args decorator on their
  constructors to get any configuration needed to construct the inner device
  instance.

  Example:
    class BleSnifferPlug(DeviceWrappingPlug):
      ...
      @conf.inject_positional_args
      def __init__(self, ble_sniffer_host, ble_sniffer_port):
        super(BleSnifferPlug, self).__init__(
            ble_sniffer.BleSniffer(ble_sniffer_host, ble_sniffer_port))
        ...

  Because not all third-party device and instrument control libraries can be
  counted on to do sufficient logging, some debug logging is provided here in
  the plug layer to show which attributes were called and with what arguments.

  Args:
    device: The device to wrap; must not be None.

  Raises:
    openhtf.plugs.InvalidPlugError: The _device attribute has the value None
        when attribute access is attempted.
  """

  verbose = True  # overwrite on subclass to disable logging_wrapper.

  def __init__(self, device):
    super(DeviceWrappingPlug, self).__init__()
    self._device = device
    if hasattr(self._device, 'tearDown') and self.uses_base_tear_down():
      self.logger.warning('Wrapped device %s implements a tearDown method, '
                          'but using the no-op BasePlug tearDown method.',
                          type(self._device))

  def __getattr__(self, attr):
    if self._device is None:
      raise openhtf.plugs.InvalidPlugError(
          'DeviceWrappingPlug instances must set the _device attribute.')
    if attr == 'as_base_types':
      return super(DeviceWrappingPlug, self).__getattr__(attr)

    attribute = getattr(self._device, attr)

    if not self.verbose or not callable(attribute):
      return attribute

    # Attribute callable; return a wrapper that logs calls with args and kwargs.
    functools.wraps(attribute, assigned=('__name__', '__doc__'))
    def logging_wrapper(*args, **kwargs):
      """Wraps a callable with a logging statement."""
      args_strings = tuple(short_repr(arg) for arg in args)
      kwargs_strings = tuple(
          ('%s=%s' % (key, short_repr(val))
           for key, val in six.iteritems(kwargs))
      )
      log_line = '%s calling "%s" on device.' % (type(self).__name__, attr)
      if args_strings or kwargs_strings:
        log_line += ' Args: \n  %s' % (', '.join(args_strings + kwargs_strings))
      self.logger.debug(log_line)
      return attribute(*args, **kwargs)

    return logging_wrapper




# Copyright 2016 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility helpers for xmlrpclib."""

import http.client
import xmlrpc.server
import socketserver
import sys
import threading
import xmlrpc.client
import collections

DEFAULT_PROXY_TIMEOUT_S = 3

# https://github.com/PythonCharmers/python-future/issues/280
if sys.version_info[0] < 3:
  from SimpleXMLRPCServer import SimpleXMLRPCServer
else:
  from xmlrpc.server import SimpleXMLRPCServer as SimpleXMLRPCServer


class TimeoutHTTPConnection(http.client.HTTPConnection):
  def __init__(self, timeout_s, *args, **kwargs):
    http.client.HTTPConnection.__init__(self, *args, **kwargs)
    self.timeout_s = timeout_s

  def settimeout(self, timeout_s):
    self.timeout_s = timeout_s
    self.sock.settimeout(self.timeout_s)

  def connect(self):
    http.client.HTTPConnection.connect(self)
    self.sock.settimeout(self.timeout_s)


class TimeoutTransport(xmlrpc.client.Transport):
  def __init__(self, timeout_s, *args, **kwargs):
    xmlrpc.client.Transport.__init__(self, *args, **kwargs)
    self._connection = None
    self.timeout_s = timeout_s

  def settimeout(self, timeout_s):
    self.timeout_s = timeout_s
    if self._connection:
      self._connection[1].settimeout(timeout_s)

  def make_connection(self, host):
    if not self._connection or host != self._connection[0]:
      self._connection = host, TimeoutHTTPConnection(self.timeout_s, host)
    return self._connection[1]


class BaseServerProxy(xmlrpc.client.ServerProxy, object):
  """New-style base class for ServerProxy, allows for use of Mixins below."""


class TimeoutProxyMixin(object):
  """Timeouts for ServerProxy objects."""
  def __init__(self, *args, **kwargs):
    super(TimeoutProxyMixin, self).__init__(
        transport=TimeoutTransport(
            kwargs.pop('timeout_s', DEFAULT_PROXY_TIMEOUT_S)),
        *args, **kwargs)

  def __settimeout(self, timeout_s):
    self.__transport.settimeout(timeout_s)


class TimeoutProxyServer(TimeoutProxyMixin, BaseServerProxy):
  """A BaseServerProxy plus timeouts."""


class LockedProxyMixin(object):
  """A ServerProxy that locks calls to methods."""
  def __init__(self, *args, **kwargs):
    super(LockedProxyMixin, self).__init__(*args, **kwargs)
    self._lock = threading.Lock()

  def __getattr__(self, attr):
    method = super(LockedProxyMixin, self).__getattr__(attr)
    if isinstance(method, collections.Callable):
      # xmlrpc doesn't support **kwargs, so only accept *args.
      def _wrapper(*args):
        with self._lock:
          return method(*args)
      # functools.wraps() doesn't work with _Method internal type within
      # xmlrpclib.  We only care about the name anyway, so manually set it.
      _wrapper.__name__ = attr
      return _wrapper
    return method


class LockedTimeoutProxy(TimeoutProxyMixin, LockedProxyMixin, BaseServerProxy):
  """ServerProxy with additional features we use."""


class SimpleThreadedXmlRpcServer(
    socketserver.ThreadingMixIn, SimpleXMLRPCServer):
  """Helper for handling multiple simultaneous RPCs in threads."""
  daemon_threads = True



# Copyright 2018 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example OpenHTF Phase Groups.

PhaseGroups are used to control phase shortcutting due to terminal errors to
better guarentee when teardown phases run.
"""

import openhtf as htf


def setup_phase(test):
  test.logger.info('Setup in a group.')


def main_phase(test):
  test.logger.info('This is a main phase.')


def teardown_phase(test):
  test.logger.info('Teardown phase.')


def inner_main_phase(test):
  test.logger.info('Inner main phase.')


def inner_teardown_phase(test):
  test.logger.info('Inner teardown phase.')


def error_setup_phase(test):
  test.logger.info('Error in setup phase.')
  return htf.PhaseResult.STOP


def error_main_phase(test):
  test.logger.info('Error in main phase.')
  return htf.PhaseResult.STOP


def run_basic_group():
  """Run the basic phase group example.

  In this example, there are no terminal phases; all phases are run.
  """
  test = htf.Test(htf.PhaseGroup(
      setup=[setup_phase],
      main=[main_phase],
      teardown=[teardown_phase],
  ))
  test.execute()


def run_setup_error_group():
  """Run the phase group example where an error occurs in a setup phase.

  The terminal setup phase shortcuts the test.  The main phases are
  skipped.  The PhaseGroup is not entered, so the teardown phases are also
  skipped.
  """
  test = htf.Test(htf.PhaseGroup(
      setup=[error_setup_phase],
      main=[main_phase],
      teardown=[teardown_phase],
  ))
  test.execute()


def run_main_error_group():
  """Run the phase group example where an error occurs in a main phase.

  The main phase in this example is terminal.  The PhaseGroup was entered
  because the setup phases ran without error, so the teardown phases are run.
  The other main phase is skipped.
  """
  test = htf.Test(htf.PhaseGroup(
      setup=[setup_phase],
      main=[error_main_phase, main_phase],
      teardown=[teardown_phase],
  ))
  test.execute()


def run_nested_groups():
  """Run the nested groups example.

  This example shows a PhaseGroup in a PhaseGroup.  No phase is terminal, so all
  are run in the order;
    main_phase
    inner_main_phase
    inner_teardown_phase
    teardown_phase
  """
  test = htf.Test(
      htf.PhaseGroup(
          main=[
              main_phase,
              htf.PhaseGroup.with_teardown(inner_teardown_phase)(
                  inner_main_phase),
          ],
          teardown=[teardown_phase]
      )
  )
  test.execute()


def run_nested_error_groups():
  """Run nested groups example where an error occurs in nested main phase.

  In this example, the first main phase in the nested PhaseGroup errors out.
  The other inner main phase is skipped, as is the outer main phase.  Both
  PhaseGroups were entered, so both teardown phases are run.
  """
  test = htf.Test(
      htf.PhaseGroup(
          main=[
              htf.PhaseGroup.with_teardown(inner_teardown_phase)(
                  error_main_phase, main_phase),
              main_phase,
          ],
          teardown=[teardown_phase],
      )
  )
  test.execute()


def run_nested_error_skip_unentered_groups():
  """Run nested groups example where an error occurs in outer main phase.

  Lastly, the first main phase in the outer PhaseGroup errors out.  This skips
  the nested PhaseGroup and the other outer main phase.  The outer PhaseGroup
  was entered, so its teardown phase runs.
  """
  test = htf.Test(
      htf.PhaseGroup(
          main=[
              error_main_phase,
              htf.PhaseGroup.with_teardown(inner_teardown_phase)(main_phase),
              main_phase,
          ],
          teardown=[teardown_phase],
      )
  )
  test.execute()


if __name__ == '__main__':
  run_basic_group()
  run_setup_error_group()
  run_main_error_group()
  run_nested_groups()
  run_nested_error_groups()
  run_nested_error_skip_unentered_groups()



import logging

from spintop_openhtf import conf, plugs

DEFAULT_FORMAT = '%(asctime)-15s: %(message)s'

class UnboundPlug(object):
    """ A generic interface base class that allows intelligent creation of OpenHTF plugs
    without limiting its usage to OpenHTF.

    Attributes:
        logger:
            If inside an OpenHTF plug, this is the OpenHTF provided logger. If not, it is 
            a logger with the object ID as name.
    """
    def __init__(self):
        super().__init__()
        if not hasattr(self, 'logger'):
            # Check conditionnaly because openhtf BasePlug will raise an
            # exception if we define self.logger.
            self.logger = logging.getLogger(str(id(self)))

    def tearDown(self):
        """Tear down the plug instance. This is part of the OpenHTF Plug contract"""
        self.close()
        self.logger.close_log()

    def open(self, *args, **kwargs):
        """Abstract method: Open resources related to this interface."""
        raise NotImplementedError()

    def close(self):
        """Abstract method: Close resources related to this interface."""
        raise NotImplementedError()

    def log_to_stream(self, stream=None, **kwargs):
        """Create and add to :attr:`logger` a StreamHandler that streams to :obj:`stream`"""
        handler = logging.StreamHandler(stream)
        self._add_handler(handler, **kwargs)

    def log_to_filename(self, filename, **kwargs):
        """Create and add to :attr:`logger` a FileHandler that logs to :obj:`filename`."""
        handler = logging.FileHandler(filename)
        self._add_handler(handler, **kwargs)

    def _add_handler(self, handler, level=logging.DEBUG, format=DEFAULT_FORMAT):
        handler.setLevel(level)
        if format:
            handler.setFormatter(logging.Formatter(format))
        self.logger.addHandler(handler)

    def close_log(self):
        """Close all :attr:`logger` handlers."""
        for handler in self.logger.handlers:
            handler.close()

    @classmethod
    def as_plug(cls, name, **kwargs_values):
        """ Create a bound plug that will retrieve values from conf or the passed values here.

        Take SSHInterface for example. 
        
        .. highlight:: python
        .. code-block:: python

            from spintop_openhtf.plugs import from_conf
            from spintop_openhtf.plugs.ssh import SSHInterface

            MySSHInterface = SSHInterface.as_plug(
                'MySSHInterface', # The name of this plug as it will appear in logs
                addr=from_conf( # from_conf will retrieve the conf value named like this.
                    'my_ssh_addr',
                    description="The addr of my device."
                ), 
                username='x', # Always the same
                password='y'  # Always the same
            )
        """
        return plug_factory(cls, name, **kwargs_values)


def plug_factory(plug_cls, name, **kwargs_values):

    for key, value in kwargs_values.items():
        if isinstance(value, from_conf):
            value.declare()
            
    def resolve(value):
        if isinstance(value, from_conf):
            return value.resolve()
        else:
            return value


    class _AnonymousPlug(plugs.BasePlug, plug_cls):
        def __init__(self):
            kwargs = {key: resolve(value) for key, value in kwargs_values.items()}
            super().__init__(**kwargs)

    _AnonymousPlug.__name__ = name
    return _AnonymousPlug

class from_conf(object):
    def __init__(self, conf_name, description=None):
        self.conf_name = conf_name
        self.description = description
    
    def resolve(self):
        return conf[self.conf_name]

    def declare(self):
        conf.declare(self.conf_name, description=self.description)


# Copyright 2016 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example OpenHTF test logic.

Run with (your virtualenv must be activated first):

  python hello_world.py

Afterwards, take a look at the hello_world.json output file.  This will
give you a basic idea of what a minimal test outputs.

For more information on measurements, see the measurements.py example.

TODO(someone): Write an output example.
For more information on output, see the output.py example.
"""

# Import openhtf with an abbreviated name, as we'll be using a bunch of stuff
# from it throughout our test scripts. See __all__ at the top of
# openhtf/__init__.py for details on what's in top-of-module namespace.
import openhtf as htf

# Import this output mechanism as it's the specific one we want to use.
from openhtf.output.callbacks import json_factory

from openhtf.plugs import user_input


# The @htf.measures annotation notifies the OpenHTF framework that this test
# phase will be taking a measurement that we'd like to call
# 'hello_world_measurement'.  Measurements can be accessed and set via
# the 'test' object, always passed as the first argument to test phases.
@htf.measures(htf.Measurement('hello_world_measurement'))
def hello_world(test):
  """A hello world test phase."""
  # At the heart of an OpenHTF test script are the test phases, such as
  # this one.  Any callable can be used as a test phase, so long as it
  # accepts a single argument that is the 'test' object.  This test object
  # is how you will interact with the OpenHTF test framework once a test is
  # running.  See other examples for more complex cases, but here is a good
  # example of the sort of minimal functionality you're likely to use.

  # The test.logger attribute is a Python logger instance that is configured
  # to log to the test record we will output at the end of the test.  This
  # is the recommended way to do any logging within test phases (this is also
  # how to get logs to show up in the frontend).
  test.logger.info('Hello World!')

  # As described above, this is how measurements are set.  The value that has
  # been set can also be accessed, but should only be set once (this will be
  # enforced in the future, for now it's best-practice).
  test.measurements.hello_world_measurement = 'Hello Again!'


if __name__ == '__main__':
  # We instantiate our OpenHTF test with the phases we want to run as args.
  # Multiple phases would be passed as additional args, and additional
  # keyword arguments may be passed as well.  See other examples for more
  # complex uses.
  test = htf.Test(hello_world)

  # In order to view the result of the test, we have to output it somewhere,
  # and a local JSON file is a convenient way to do this.  Custom output
  # mechanisms can be implemented, but for now we'll just keep it simple.
  # This will always output to the same ./hello_world.json file, formatted
  # slightly for human readability.
  test.add_output_callbacks(
      json_factory.OutputToJSON('./{dut_id}.hello_world.json', indent=2))

  # prompt_for_test_start prompts the operator for a DUT ID, a unique identifier
  # for the DUT (Device Under Test).  OpenHTF requires that a DUT ID is set
  # each time a test is executed.  It may be set programmatically, but the
  # simplest way to get one is to prompt the user for it.  If test_start is
  # not provided, the test will start immediately and assume the DUT ID will
  # be set later (OpenHTF will raise an exception when the test completes if
  # a DUT ID has not been set).
  test.execute(test_start=user_input.prompt_for_test_start())



"""Output and/or upload a TestRun or MfgEvent proto for mfg-inspector.com.
"""

import json
import logging
import threading
import time
import zlib

import httplib2
import oauth2client.client

from openhtf.output import callbacks
from openhtf.output.proto import guzzle_pb2
from openhtf.output.proto import test_runs_converter

import six
from six.moves import range


class UploadFailedError(Exception):
  """Raised when an upload to mfg-inspector fails."""


class InvalidTestRunError(Exception):
  """Raised if test run is invalid."""


def _send_mfg_inspector_request(envelope_data, credentials, destination_url):
  """Send upload http request.  Intended to be run in retry loop."""
  logging.info('Uploading result...')
  http = httplib2.Http()

  if credentials.access_token_expired:
    credentials.refresh(http)
  credentials.authorize(http)

  resp, content = http.request(destination_url, 'POST', envelope_data)

  try:
    result = json.loads(content)
  except Exception:
    logging.debug('Upload failed with response %s: %s', resp, content)
    raise UploadFailedError(resp, content)

  if resp.status != 200:
    logging.debug('Upload failed: %s', result)
    raise UploadFailedError(result['error'], result)

  return result


def send_mfg_inspector_data(inspector_proto, credentials, destination_url):
  """Upload MfgEvent to steam_engine."""
  envelope = guzzle_pb2.TestRunEnvelope()
  envelope.payload = zlib.compress(inspector_proto.SerializeToString())
  envelope.payload_type = guzzle_pb2.COMPRESSED_MFG_EVENT
  envelope_data = envelope.SerializeToString()

  for _ in range(5):
    try:
      result = _send_mfg_inspector_request(
          envelope_data, credentials, destination_url)
      return result
    except UploadFailedError:
      time.sleep(1)

  logging.critical(
      'Could not upload to mfg-inspector after 5 attempts. Giving up.')

  return {}


class _MemStorage(oauth2client.client.Storage):
  # pylint: disable=invalid-name
  """Helper Storage class that keeps credentials in memory."""

  def __init__(self):
    self._lock = threading.Lock()
    self._credentials = None

  def acquire_lock(self):
    self._lock.acquire(True)

  def release_lock(self):
    self._lock.release()

  def locked_get(self):
    return self._credentials

  def locked_put(self, credentials):
    self._credentials = credentials


class MfgInspector(object):
  """Interface to convert a TestRun to a mfg-inspector compatible proto.

  Instances of this class are typically used to create callbacks that are
  compatible with the OpenHTF output callbacks.

  Typical usage:
  interface = mfg_inspector.MfgInspector.from_json().set_converter(
    my_custom_converter)
  my_tester.add_output_callbacks(interface.save_to_disk(), interface.upload())

  **Important** the conversion of the TestRecord to protofbuf as specified in
  the _converter callable attribute only occurs once and the resulting protobuf
  is cached in memory on the instance.

  The upload callback will upload to mfg-inspector.com using the given
  username and authentication key (which should be the key data itself, not a
  filename or file).

  In typical productin setups, we *first* save the protobuf to disk then attempt
  to upload the protobuf to mfg-inspector.  In the event of a network outage,
  the result of the test run is available on disk and a separate process can
  retry the upload when network is available.
  """

  TOKEN_URI = 'https://accounts.google.com/o/oauth2/token'
  SCOPE_CODE_URI = 'https://www.googleapis.com/auth/glass.infra.quantum_upload'
  DESTINATION_URL = ('https://clients2.google.com/factoryfactory/'
                     'uploads/quantum_upload/?json')

  # These attributes control format of callback and what actions are undertaken
  # when called.  These should either be set by a subclass or via configure.

  # _converter is a callable that can be set either via set_converter method
  # or by defining a _converter @staticmethod on subclasses.
  _converter = None

  # A default filename pattern can be specified on subclasses for use when
  # saving to disk via save_to_disk.
  _default_filename_pattern = None

  def __init__(self, user=None, keydata=None,
               token_uri=TOKEN_URI, destination_url=DESTINATION_URL):
    self.user = user
    self.keydata = keydata
    self.token_uri = token_uri
    self.destination_url = destination_url

    if user and keydata:
      self.credentials = oauth2client.client.SignedJwtAssertionCredentials(
          service_account_name=self.user,
          private_key=six.ensure_binary(self.keydata),
          scope=self.SCOPE_CODE_URI,
          user_agent='OpenHTF Guzzle Upload Client',
          token_uri=self.token_uri)
      self.credentials.set_store(_MemStorage())
    else:
      self.credentials = None

    self.upload_result = None

    self._cached_proto = None

  @classmethod
  def from_json(cls, json_data):
    """Create an uploader given (parsed) JSON data.

    Note that this is a JSON-formatted key file downloaded from Google when
    the service account key is created, *NOT* a json-encoded
    oauth2client.client.SignedJwtAssertionCredentials object.

    Args:
      json_data: Dict containing the loaded JSON key data.

    Returns:
      a MfgInspectorCallback with credentials.
    """
    return cls(user=json_data['client_email'],
               keydata=json_data['private_key'],
               token_uri=json_data['token_uri'])

  def _convert(self, test_record_obj):
    """Convert and cache a test record to a mfg-inspector proto."""

    if self._cached_proto is None:
      self._cached_proto = self._converter(test_record_obj)

    return self._cached_proto

  def save_to_disk(self, filename_pattern=None):
    """Returns a callback to convert test record to proto and save to disk."""
    if not self._converter:
      raise RuntimeError(
          'Must set _converter on subclass or via set_converter before calling '
          'save_to_disk.')

    pattern = filename_pattern or self._default_filename_pattern
    if not pattern:
      raise RuntimeError(
          'Must specify provide a filename_pattern or set a '
          '_default_filename_pattern on subclass.')

    def save_to_disk_callback(test_record_obj):
      proto = self._convert(test_record_obj)
      output_to_file = callbacks.OutputToFile(pattern)

      with output_to_file.open_output_file(test_record_obj) as outfile:
        outfile.write(proto.SerializeToString())

    return save_to_disk_callback

  def upload(self):
    """Returns a callback to convert a test record to a proto and upload."""
    if not self._converter:
      raise RuntimeError(
          'Must set _converter on subclass or via set_converter before calling '
          'upload.')

    if not self.credentials:
      raise RuntimeError('Must provide credentials to use upload callback.')

    def upload_callback(test_record_obj):
      proto = self._convert(test_record_obj)
      self.upload_result = send_mfg_inspector_data(
          proto, self.credentials, self.destination_url)

    return upload_callback

  def set_converter(self, converter):
    """Set converter callable to convert a OpenHTF tester_record to a proto.

    Args:
      converter: a callable that accepts an OpenHTF TestRecord and returns a
        manufacturing-inspector compatible protobuf.

    Returns:
      self to make this call chainable.
    """
    assert callable(converter), 'Converter must be callable.'

    self._converter = converter

    return self


# LEGACY / DEPRECATED
class UploadToMfgInspector(MfgInspector):
  """Generate a mfg-inspector TestRun proto and upload it.

  LEGACY / DEPRECATED
  This class is provided only for legacy reasons and may be deleted in future.
  Please replace usage by configuring a MfgInspectorCallback directly. For
  example:
  test.add_output_callbacks(
    mfg_inspector.MfgInspectorCallback.from_json(**json_data).set_converter(
      test_runs_converter.test_run_from_test_record).upload()
  )
  """

  @staticmethod
  def _converter(test_record_obj):
    return test_runs_converter.test_run_from_test_record(test_record_obj)

  def __call__(self, test_record_obj):  # pylint: disable=invalid-name
    upload_callback = self.upload()
    upload_callback(test_record_obj)




import pytest
try:
    import tornado
    TORNADO_AVAILABLE = True
except ImportError:
    TORNADO_AVAILABLE = False
    
from spintop_openhtf import TestPlan, PhaseResult

A_TEST_PLAN_NAME = 'test-plan'
A_TEST_NAME = 'Test1'

@pytest.fixture()
def test_plan():
    return TestPlan(A_TEST_PLAN_NAME, store_result=False)

def test_simple_testcase(test_plan):
    
    @test_plan.testcase(A_TEST_NAME)
    def test_1(test): # pylint: disable=unused-variable
        pass
    
    assert test_plan.phase_group.main and test_plan.phase_group.main[0].name == A_TEST_NAME

@pytest.mark.skipif(not TORNADO_AVAILABLE, reason="Requires the GUI extras [server]")
def test_no_trigger_and_run_with_gui(test_plan):
    test_plan.no_trigger()
    executed = []
    
    assert not executed
    
    @test_plan.testcase(A_TEST_NAME)
    def test_1(test): # pylint: disable=unused-variable
        executed.append(True)
    
    test_plan.run_once(launch_browser=False)
    
    assert executed
    
def test_custom_trigger(test_plan):
    executed = []
    
    assert not executed
    
    @test_plan.trigger('trigger')
    def test_trigger(test):
        executed.append(True)
    
    @test_plan.testcase(A_TEST_NAME)
    def test_1(test): # pylint: disable=unused-variable
        executed.append(True)
    
    test_plan.execute()
    
    assert len(executed) == 2
    
def test_setup_fn(test_plan):
    test_plan.no_trigger()
    executed = []
    
    assert not executed
    
    @test_plan.setup('fail-setup')
    def test_fail_setup(test):
        executed.append(True)
        return PhaseResult.STOP 
    
    @test_plan.testcase(A_TEST_NAME)
    def test_1(test): # pylint: disable=unused-variable
        executed.append(True)
    
    test_plan.execute()
    
    assert len(executed) == 1
    
def test_sub_sequence(test_plan):
    test_plan.no_trigger()
    executed = []
    
    def my_test(test):
        executed.append(True)
        
    test_plan.testcase('test1')(my_test) # 1
    subseq = test_plan.sub_sequence('subseq')
    subseq.testcase('test2')(my_test) # 2
    subseq.testcase('test3')(my_test) # 3
    
    test_plan.execute()
    assert len(executed) == 3
        

def test_run_console():
    plan = TestPlan('empty')

    @plan.testcase('Nothing')
    def test_something(test):
        pass

    plan.no_trigger()
    plan.run_console(once=True)
    
    
    
    


# Copyright 2016 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for automic_write a new file."""

import os
import tempfile
from contextlib import contextmanager

@contextmanager
def atomic_write(filename, filesync=False):
  """ atomically write a file (using a temporary file).

  filename: the file to be written
  filesync: flush the file to disk
  """

  tmpf = tempfile.NamedTemporaryFile(delete=False)
  try:
    with open(tmpf.name, 'w') as curfile:
      yield curfile
      if filesync:
        curfile.flush()
        os.fsync(curfile.fileno())
    os.rename(tmpf.name, filename)
  finally:
    try:
      os.remove(tmpf.name)
    except (IOError, OSError):
      pass



# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""A simple utility to do timeout checking."""

import contextlib
import functools
import logging
import threading
import time

_LOG = logging.getLogger(__name__)

class PolledTimeout(object):
  """An object which tracks if a timeout has expired."""

  def __init__(self, timeout_s):
    """Construct a PolledTimeout object.

    Args:
      timeout_s: This may either be a number or None. If a number, this object
        will consider to be expired after number seconds after construction. If
        None, this object never expires.
    """
    self.start = time.time()
    self.timeout_s = timeout_s

  @classmethod
  def from_millis(cls, timeout_ms):
    """Create a new PolledTimeout if needed.

    If timeout_ms is already a PolledTimeout, just return it, otherwise create a
    new PolledTimeout with the given timeout in milliseconds.

    Args:
      timeout_ms: PolledTimeout object, or number of milliseconds to use for
        creating a new one.

    Returns:
      A PolledTimeout object that will expire in timeout_ms milliseconds, which
    may be timeout_ms itself, or a newly allocated PolledTimeout.
    """
    if hasattr(timeout_ms, 'has_expired'):
      return timeout_ms
    if timeout_ms is None:
      return cls(None)
    return cls(timeout_ms / 1000.0)

  @classmethod
  def from_seconds(cls, timeout_s):
    """Create a new PolledTimeout if needed.

    If timeout_s is already a PolledTimeout, just return it, otherwise create a
    new PolledTimeout with the given timeout in seconds.

    Args:
      timeout_s: PolledTimeout object, or number of seconds to use for creating
    a new one.

    Returns:
      A PolledTimeout object that will expire in timeout_s seconds, which may
    be timeout_s itself, or a newly allocated PolledTimeout.
    """
    if hasattr(timeout_s, 'has_expired'):
      return timeout_s
    return cls(timeout_s)

  def restart(self):
    """Restarts the timeout. Initializing the start time to now."""
    self.start = time.time()

  def expire(self):
    """Expire the timeout immediately."""
    self.timeout_s = 0

  def has_expired(self):
    """Returns True if the timeout has expired."""
    if self.timeout_s is None:
      return False
    return self.seconds >= self.timeout_s

  # Bad API. alusco is sometimes bad at naming.
  Poll = has_expired  # pylint: disable=invalid-name

  # pylint: disable=missing-docstring
  @property
  def seconds(self):
    return time.time() - self.start

  @property
  def remaining(self):
    if self.timeout_s is None:
      return None
    # We max() against 0 to ensure we don't return a (slightly) negative number.
    # This reduces races between .has_expired() calls and sleeping/waiting
    # .remaining seconds.
    return max(0, self.timeout_s - self.seconds)

  @property
  def remaining_ms(self):
    if self.timeout_s is None:
      return None
    return self.remaining * 1000

  # pylint: disable=missing-docstring


# There's now no way to tell if a timeout occurred generically
# which sort of sucks (for generic validation fn)
def loop_until_timeout_or_valid(timeout_s, function, validation_fn, sleep_s=1):  # pylint: disable=invalid-name
  """Loops until the specified function returns valid or a timeout is reached.

  Note: The function may return anything which, when passed to validation_fn,
  evaluates to implicit True.  This function will loop calling the function as
  long as the result of validation_fn(function_result) returns something which
  evaluates to False. We ensure function is called at least once regardless
  of timeout.

  Args:
    timeout_s: The number of seconds to wait until a timeout condition is
        reached. As a convenience, this accepts None to mean never timeout.  Can
        also be passed a PolledTimeout object instead of an integer.
    function: The function to call each iteration.
    validation_fn: The validation function called on the function result to
        determine whether to keep looping.
    sleep_s: The number of seconds to wait after calling the function.

  Returns:
    Whatever the function returned last.
  """
  if timeout_s is None or not hasattr(timeout_s, 'has_expired'):
    timeout_s = PolledTimeout(timeout_s)

  while True:
    # Calls the function at least once
    result = function()
    if validation_fn(result) or timeout_s.has_expired():
      return result
    time.sleep(sleep_s)


def loop_until_timeout_or_true(timeout_s, function, sleep_s=1):  # pylint: disable=invalid-name
  """Loops until the specified function returns True or a timeout is reached.

  Note: The function may return anything which evaluates to implicit True.  This
  function will loop calling it as long as it continues to return something
  which evaluates to False.  We ensure this method is called at least once
  regardless of timeout.

  Args:
    timeout_s: The number of seconds to wait until a timeout condition is
        reached. As a convenience, this accepts None to mean never timeout.  Can
        also be passed a PolledTimeout object instead of an integer.
    function: The function to call each iteration.
    sleep_s: The number of seconds to wait after calling the function.

  Returns:
    Whatever the function returned last.
  """
  return loop_until_timeout_or_valid(timeout_s, function, lambda x: x, sleep_s)


def loop_until_timeout_or_not_none(timeout_s, function, sleep_s=1):  # pylint: disable=invalid-name
  """Loops until the specified function returns non-None or until a timeout.

  Args:
    timeout_s: The number of seconds to wait until a timeout condition is
        reached. As a convenience, this accepts None to mean never timeout.  Can
        also be passed a PolledTimeout object instead of an integer.
    function: The function to call each iteration.
    sleep_s: The number of seconds to wait after calling the function.

  Returns:
    Whatever the function returned last.
  """
  return loop_until_timeout_or_valid(
      timeout_s, function, lambda x: x is not None, sleep_s)


def loop_until_true_else_raise(timeout_s,
                               function,
                               invert=False,
                               message=None,
                               sleep_s=1):
  """Repeatedly call the given function until truthy, or raise on a timeout.

  Args:
    timeout_s: The number of seconds to wait until a timeout condition is
        reached. As a convenience, this accepts None to mean never timeout. Can
        also be passed a PolledTimeout object instead of an integer.
    function: The function to call each iteration.
    invert: If True, wait for the callable to return falsey instead of truthy.
    message: Optional custom error message to use on a timeout.
    sleep_s: Seconds to sleep between call attempts.

  Returns:
    The final return value of the function.

  Raises:
    RuntimeError if the timeout is reached before the function returns truthy.
  """
  def validate(x):
    return bool(x) != invert

  result = loop_until_timeout_or_valid(timeout_s, function, validate, sleep_s=1)
  if validate(result):
    return result

  if message is not None:
    raise RuntimeError(message)

  name = '(unknown)'
  if hasattr(function, '__name__'):
    name = function.__name__
  elif (isinstance(function, functools.partial)
        and hasattr(function.func, '__name__')):
    name = function.func.__name__
  raise RuntimeError(
      'Function %s failed to return %s within %d seconds.'
      % (name, 'falsey' if invert else 'truthy', timeout_s)) 


class Interval(object):
  """An object which can execute a method on an interval."""

  def __init__(self, method, stop_if_false=False):
    """Initializes the Interval.

    Args:
      method: A callable to execute, it should take no arguments.
      stop_if_false: If True, the interval will exit if the method returns
      False.
    """
    self.method = method
    self.stopped = threading.Event()
    self.thread = None
    self.stop_if_false = stop_if_false

  @property
  def running(self):
    if self.thread:
      return self.thread.isAlive()
    return False

  def start(self, interval_s):
    """Starts executing the method at the specified interval.

    Args:
      interval_s: The amount of time between executions of the method.
    Returns:
      False if the interval was already running.
    """
    if self.running:
      return False

    self.stopped.clear()

    def _execute():
      # Always execute immediately once
      if not self.method() and self.stop_if_false:
        return
      while not self.stopped.wait(interval_s):
        if not self.method() and self.stop_if_false:
          return

    self.thread = threading.Thread(target=_execute)
    self.thread.daemon = True
    self.thread.start()
    return True

  def stop(self, timeout_s=None):
    """Stops the interval.

    If a timeout is provided and stop returns False then the thread is
    effectively abandoned in whatever state it was in (presumably dead-locked).

    Args:
      timeout_s: The time in seconds to wait on the thread to finish.  By
          default it's forever.
    Returns:
      False if a timeout was provided and we timed out.
    """
    self.stopped.set()
    if self.thread:
      self.thread.join(timeout_s)
      return not self.thread.isAlive()
    else:
      return True

  def join(self, timeout_s=None):
    """Joins blocking until the interval ends or until timeout is reached.

    Args:
      timeout_s: The time in seconds to wait, defaults to forever.
    Returns:
      True if the interval is still running and we reached the timeout.
    """
    if not self.thread:
      return False
    self.thread.join(timeout_s)
    return self.running


def execute_forever(method, interval_s):  # pylint: disable=invalid-name
  """Executes a method forever at the specified interval.

  Args:
    method: The callable to execute.
    interval_s: The number of seconds to start the execution after each method
        finishes.
  Returns:
    An Interval object.
  """
  interval = Interval(method)
  interval.start(interval_s)
  return interval


def execute_until_false(method, interval_s):  # pylint: disable=invalid-name
  """Executes a method forever until the method returns a false value.

  Args:
    method: The callable to execute.
    interval_s: The number of seconds to start the execution after each method
        finishes.
  Returns:
    An Interval object.
  """
  interval = Interval(method, stop_if_false=True)
  interval.start(interval_s)
  return interval


# pylint: disable=invalid-name
def retry_until_true_or_limit_reached(method, limit, sleep_s=1,
                                      catch_exceptions=()):
  """Executes a method until the retry limit is hit or True is returned."""
  return retry_until_valid_or_limit_reached(
      method, limit, lambda x: x, sleep_s, catch_exceptions)


def retry_until_not_none_or_limit_reached(method, limit, sleep_s=1,
                                          catch_exceptions=()):
  """Executes a method until the retry limit is hit or not None is returned."""
  return retry_until_valid_or_limit_reached(
      method, limit, lambda x: x is not None, sleep_s, catch_exceptions)


def retry_until_valid_or_limit_reached(method, limit, validation_fn, sleep_s=1,
                                       catch_exceptions=()):
  """Executes a method until the retry limit or validation_fn returns True.

  The method is always called once so the effective lower limit for 'limit' is
  1.  Passing in a number less than 1 will still result it the method being
  called once.

  Args:
    method: The method to execute should take no arguments.
    limit: The number of times to try this method.  Must be >0.
    validation_fn: The validation function called on the function result to
        determine whether to keep looping.
    sleep_s: The time to sleep in between invocations.
    catch_exceptions: Tuple of exception types to catch and count as failures.
  Returns:
      Whatever the method last returned, implicit False would indicate the
        method never succeeded.
  """
  assert limit > 0, 'Limit must be greater than 0'

  def _execute_method(helper):
    try:
      return method()
    except catch_exceptions:
      if not helper.remaining:
        raise
      return None

  helper = RetryHelper(limit - 1)
  result = _execute_method(helper)
  while not validation_fn(result) and helper.retry_if_possible():
    time.sleep(sleep_s)
    result = _execute_method(helper)
  return result

# pylint: disable=invalid-name


@contextlib.contextmanager
def take_at_least_n_seconds(time_s):
  """A context manager which ensures it takes at least time_s to execute.

  Example:
    with take_at_least_n_seconds(5):
      do.Something()
      do.SomethingElse()
    # if Something and SomethingElse took 3 seconds, the with block with sleep
    # for 2 seconds before exiting.
  Args:
    time_s: The number of seconds this block should take.  If it doesn't take at
      least this time, then this method blocks during __exit__.
  Yields:
    To do some actions then on completion waits the remaining time.
  """
  timeout = PolledTimeout(time_s)
  yield
  while not timeout.has_expired():
    time.sleep(timeout.remaining)


def take_at_most_n_seconds(time_s, func, *args, **kwargs):
  """A function that returns whether a function call took less than time_s.

  NOTE: The function call is not killed and will run indefinitely if hung.

  Args:
    time_s: Maximum amount of time to take.
    func: Function to call.
    *args: Arguments to call the function with.
    **kwargs: Keyword arguments to call the function with.
  Returns:
    True if the function finished in less than time_s seconds.
  """
  thread = threading.Thread(target=func, args=args, kwargs=kwargs)
  thread.start()
  thread.join(time_s)
  if thread.is_alive():
    return False
  return True


def execute_after_delay(time_s, func, *args, **kwargs):
  """A function that executes the given function after a delay.

  Executes func in a separate thread after a delay, so that this function
  returns immediately.  Note that any exceptions raised by func will be
  ignored (but logged).  Also, if time_s is a PolledTimeout with no expiration,
  then this method simply returns immediately and does nothing.

  Args:
    time_s: Delay in seconds to wait before executing func, may be a
      PolledTimeout object.
    func: Function to call.
    *args: Arguments to call the function with.
    **kwargs: Keyword arguments to call the function with.
  """
  timeout = PolledTimeout.from_seconds(time_s)

  def target():
    time.sleep(timeout.remaining)
    try:
      func(*args, **kwargs)
    except Exception:  # pylint: disable=broad-except
      _LOG.exception('Error executing %s after %s expires.', func, timeout)
  if timeout.remaining is not None:
    thread = threading.Thread(target=target)
    thread.start()


class RetryHelper(object):
  """A helper with to simplify retrying.

  Attributes:
    remaining: The remaining number of retries.
  """

  def __init__(self, retries):
    """Initializes this object.

    Args:
      retries: The number of retries to allow.
    """
    self.remaining = retries

  def retry_if_possible(self):
    """Decrements a retry.

    Returns:
      True if you should proceed, or False if you're out of retries.
    """
    self.remaining -= 1
    return self.remaining >= 0



from uuid import uuid4
from contextlib import contextmanager

TEMPFILES_PATH = '/tempfiles/'

class TemporaryFileProvider(object):
    def __init__(self):
        self.files = {}
        self.prefix = TEMPFILES_PATH
    
    def set_prefix(self, prefix):
        self.prefix = prefix
    
    def create_url(self, filename):
        subpath = self.prefix + str(uuid4())
        self.files[subpath] = filename
        return subpath
    
    def delete_url(self, subpath):
        del self.files[subpath]
    
    @contextmanager
    def temp_file_url(self, filename):
        try:
            subpath = self.create_url(filename)
            yield subpath
        finally:
            self.delete_url(subpath)


# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""One-off utilities."""

import logging
import re
import threading
import time
import weakref

import mutablerecords

import six


def _log_every_n_to_logger(n, logger, level, message, *args):  # pylint: disable=invalid-name
  """Logs the given message every n calls to a logger.

  Args:
    n: Number of calls before logging.
    logger: The logger to which to log.
    level: The logging level (e.g. logging.INFO).
    message: A message to log
    *args: Any format args for the message.
  Returns:
    A method that logs and returns True every n calls.
  """
  logger = logger or logging.getLogger()
  def _gen():  # pylint: disable=missing-docstring
    while True:
      for _ in range(n):
        yield False
      logger.log(level, message, *args)
      yield True
  gen = _gen()
  return lambda: six.next(gen)


def log_every_n(n, level, message, *args):  # pylint: disable=invalid-name
  """Logs a message every n calls. See _log_every_n_to_logger."""
  return _log_every_n_to_logger(n, None, level, message, *args)


def time_millis():  # pylint: disable=invalid-name
  """The time in milliseconds."""
  return int(time.time() * 1000)


class NonLocalResult(mutablerecords.Record('NonLocal', [], {'result': None})):
  """Holds a single result as a nonlocal variable.

  Comparable to using Python 3's nonlocal keyword, it allows an inner function
  to set the value in an outer function's namespace:

  def WrappingFunction():
    x = NonLocalResult()
    def InnerFunction():
      # This is what we'd do in Python 3:
      # nonlocal x
      # x = 1
      # In Python 2 we use NonLocalResult instead.
      x.result = 1
    InnerFunction()
    return x.result
  """


# TODO(jethier): Add a pylint plugin to avoid the no-self-argument for this.
class classproperty(object):
  """Exactly what it sounds like.

  Note that classproperties don't have setters, so setting them will replace
  the classproperty with the new value. In most use cases (forcing subclasses
  to override the classproperty, for example) this is desired.
  """
  def __init__(self, func):
    self._func = func

  def __get__(self, instance, owner):
    return self._func(owner)


def partial_format(target, **kwargs):
  """Formats a string without requiring all values to be present.

  This function allows substitutions to be gradually made in several steps
  rather than all at once.  Similar to string.Template.safe_substitute.
  """
  output = target[:]

  for tag, var in re.findall(r'(\{(.*?)\})', output):
    root = var.split('.')[0]  # dot notation
    root = root.split('[')[0]  # dict notation
    if root in kwargs:
      output = output.replace(tag, tag.format(**{root: kwargs[root]}))

  return output

def format_string(target, kwargs):
  """Formats a string in any of three ways (or not at all).

  Args:
    target: The target string to format. This can be a function that takes a
        dict as its only argument, a string with {}- or %-based formatting, or
        a basic string with none of those. In the latter case, the string is
        returned as-is, but in all other cases the string is formatted (or the
        callback called) with the given kwargs.
        If this is None (or otherwise falsey), it is returned immediately.
    kwargs: The arguments to use for formatting.
        Passed to safe_format, %, or target if it's
        callable.
  """
  if not target:
    return target
  if callable(target):
    return target(**kwargs)
  if not isinstance(target, six.string_types):
    return target
  if '{' in target:
    return partial_format(target, **kwargs)
  if '%' in target:
    return target % kwargs
  return target


class SubscribableStateMixin(object):
  """Gives an object the capability of notifying watchers of state changes.

  The state should be represented as a dictionary and returned by _asdict.
  An object that wants to watch this object's state should call
  asdict_with_event to get the current state and an event object. This object
  can then notify watchers holding those events that the state has changed by
  calling notify_update.
  """

  def __init__(self):
    super(SubscribableStateMixin, self).__init__()
    self._lock = threading.Lock()
    self._update_events = weakref.WeakSet()

  def _asdict(self):
    raise NotImplementedError(
        'Subclasses of SubscribableStateMixin must implement _asdict.')

  def asdict_with_event(self):
    """Get a dict representation of this object and an update event.

    Returns:
      state: Dict representation of this object.
      update_event: An event that is guaranteed to be set if an update has been
          triggered since the returned dict was generated.
    """
    event = threading.Event()
    with self._lock:
      self._update_events.add(event)
    return self._asdict(), event

  def notify_update(self):
    """Notify any update events that there was an update."""
    with self._lock:
      for event in self._update_events:
        event.set()
      self._update_events.clear()



import time
import threading
from queue import Queue, Empty
from asyncio import Protocol

from io import StringIO

from ..base import UnboundPlug

class IOTargetTimeout(Exception):
    pass

class IOInterface(UnboundPlug, Protocol):
    """An interface to a read/write interface.
    
    Allows reading and writing. A background thread reads any data that comes in 
    and those lines can be accessed using the `next_line` function.
    
    """

    def __init__(self):
        super().__init__()
        self._line_buffer = ""
        self._read_lines = Queue()
        self._timeout_timer = None
        self._last_receive_time = None

        self.timeout = 0.5
        self.eol = "\n"

    def tearDown(self):
        """Tear down the plug instance."""
        self.close()
        self.logger.close_log()

    def open(self, *args, **kwargs):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

    def write(self, string):
        """Write the string into the io interface."""
        self.logger.debug("> {!r}".format(string))
        return self._write(string)

    def _write(self, string):
        raise NotImplementedError()

    def data_received(self, data):
        if self._timeout_timer is not None:
            self._timeout_timer.cancel()
        
        force_clear = False
        if self.check_receive_delta() and data is None and self._line_buffer:
            force_clear = True

        data_to_queue = None

        if not data:
            data = b''

        string = data.decode('utf8')
        if force_clear or string.endswith(self.eol) or (self._line_buffer + string).endswith(self.eol):
            data_to_queue = self._line_buffer + string
            self._line_buffer = ""
        else:
            self._line_buffer += string
        
        self._timeout_timer = threading.Timer(self.timeout, self.no_data_received)
        self._timeout_timer.start()

        if data_to_queue:
            strings_with_eol = data_to_queue.split(self.eol)
            # Last has no eol, all others do.
            for index, line in enumerate(strings_with_eol):
                if index < len(strings_with_eol) - 1:
                    line = line + self.eol

                if line:
                    self.logger.debug(line)
                    self._read_lines.put(line)

    def no_data_received(self):
        return self.data_received(None)

    def check_receive_delta(self):
        if self._last_receive_time:
            delta = time.time() - self._last_receive_time
        else:
            delta = 0

        self._last_receive_time = time.time()

        return delta > self.timeout

    def keep_lines(self, lines_to_keep):
        """Clear all lines in the buffer except the last lines_to_keep lines."""
        current_lines = self._read_lines.qsize()
        while current_lines > lines_to_keep:
            current_lines = current_lines - 1
            self.next_line()
            current_lines = self._read_lines.qsize()
        
    def clear_lines(self):
        """Clear all lines in the buffer."""
        self._read_lines.queue.clear()

    def next_line(self, timeout=10):
        """ Waits up to timeout seconds and return the next line available in the buffer.
        """
        try:
            return self._read_lines.get(timeout=timeout)
        except Empty:
            return None

    def eof_received(self):
        pass

    
    def message_target(self, message, target, timeout=None, keeplines=0, _timeout_raises=True):
        """Sends the message string and waits for any string in target to be received. 
        
        Parameters:
            message: The string to write into the io interface
            target: A string or a list of string to check for in the read lines.
            timeout: (default=None, no timeout) Wait up to this seconds for the targets. If busted, this will raise a IOTargetTimeout.
            keeplines: (default=0, discard all) Before sending the message, call self.keep_lines with this value.
            _timeout_raises: (default=True) If True, a timeout will raise an error.
        """
        self.keep_lines(keeplines)
        self.write(message)

        if isinstance(target, str):
            targets = [target]
        else:
            targets = target

        start_time = time.time()
        target_found = None
        contentstring = StringIO()

        while not target_found and (timeout is None or time.time() - start_time < timeout):
            line = self.next_line()
            contentstring.write(line)
            for targetstr in targets:
                if targetstr in line:
                    target_found = targetstr
                    break

        if target_found is None:
            log_message = 'Message Target: timeout occured while waiting for %s' % str(targets)
            if _timeout_raises:
                raise IOTargetTimeout(log_message)
        else:
            log_message = 'Message Target: Found %s' % target_found

        contentlog = contentstring.getvalue()
        contentstring.close()

        return contentlog
    
    def execute_command(self, command=None, timeout=None, target=None):
        """Adds the self.eol to command and call message_target using either
        target as target or self._target if defined. This is used for executing
        commands in a shell-like environment and to wait for the prompt.
        self._target or target should be the expected prompt."""
        if target is None:
            target = getattr(self, '_target', '$')
        command = command + self.eol
        return_value = self.message_target(command, target, timeout)
        return return_value



# Copyright 2016 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Multicast facilities for sending and receiving messages.

This module includes both a MulticastListener that listens on a multicast
socket and invokes a callback function for each message received, and a send()
function that is used to send one-shot messages to a multicast socket.
"""

import logging
import socket
import ipaddress
import struct
import sys
import threading
import time
from six.moves import queue

_LOG = logging.getLogger(__name__)

# The multicast logs can occur frequently, e.g. every couple seconds. Prevent
# most logs from ending up in the test record by default.
_LOG.setLevel(logging.WARNING)

DEFAULT_ADDRESS = '239.1.1.1'
DEFAULT_PORT = 10000
DEFAULT_TTL = 1
LOCALHOST_ADDRESS = 0x7f000001  # 127.0.0.1
MAX_MESSAGE_BYTES = 1024  # Maximum allowable message length in bytes.

def get_all_interfaces():
  if sys.platform.startswith('darwin') or sys.platform.startswith('win'):
    # Does not support socket.INADDR_ANY. List all interfaces manually.
    all_interfaces = socket.gethostbyname_ex(socket.gethostname())[2]
    return tuple(ip2int(iface) for iface in all_interfaces)
  else:
    return (socket.INADDR_ANY,)

def ip2int(addr):
    return struct.unpack("!I", socket.inet_aton(addr))[0]

def platform_multicast_bind_addr(multicast_address):
  if sys.platform.startswith('darwin') or sys.platform.startswith('win'):
    return ''
  else:
    return multicast_address

class MulticastListener(threading.Thread):
  """Agent that listens (and responds) to short messages on a multicast socket.

  The listener will force-bind to the multicast port via the SO_REUSEADDR
  option, so it's possible for multiple listeners to bind to the same port.

  Args:
    callback: A callable to invoke upon receipt of a multicast message. Will be
              called with one argument -- the text of the message received.
              callback can optionally return a string response, which will be
              transmitted back to the sender.
    address: Multicast IP address component of the socket to listen on.
    port: Multicast UDP port component of the socket to listen on.
    ttl: TTL for multicast messages. 1 to keep traffic in-network.
  """
  LISTEN_TIMEOUT_S = 60  # Seconds to listen before retrying.
  daemon = True

  def __init__(self,
               callback,
               address=DEFAULT_ADDRESS,
               port=DEFAULT_PORT,
               ttl=DEFAULT_TTL):
    super(MulticastListener, self).__init__()
    self.address = address
    self.port = port
    self.ttl = ttl
    self._callback = callback
    self._live = False
    self._sock = None

  def stop(self, timeout_s=None):
    """Stop listening for messages."""
    self._live = False
    if self._sock is not None:
      try:
        self._sock.shutdown(socket.SHUT_RDWR)
        self._sock.close()
      except socket.error:
        pass
    self.join(timeout_s)

  def _init_sock(self):
    self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    self._sock.setsockopt(socket.IPPROTO_IP,
                          socket.IP_MULTICAST_TTL,
                          self.ttl)
    self._sock.settimeout(self.LISTEN_TIMEOUT_S)

    for interface_ip in  get_all_interfaces() + (LOCALHOST_ADDRESS,):
      self._sock.setsockopt(
          socket.IPPROTO_IP,
          socket.IP_ADD_MEMBERSHIP,
          # IP_ADD_MEMBERSHIP takes the 8-byte group address followed by the IP
          # assigned to the interface on which to listen.
          struct.pack('!4sL', socket.inet_aton(self.address), interface_ip))


    self._sock.bind((platform_multicast_bind_addr(self.address), self.port))

  def _try_init(self, retries, sleep_time=2):
    while True:
      try:
        self._init_sock()
        return
      except OSError:
        retries = retries - 1
        if retries == 0:
          raise
        else:
          time.sleep(sleep_time)

  def run(self):
    """Listen for pings until stopped."""
    self._live = True
    self._try_init(20)

    while self._live:
      try:
        data, address = self._sock.recvfrom(MAX_MESSAGE_BYTES)
        data = data.decode('utf-8')
        log_line = 'Received multicast message from %s: %s' % (address, data)
        response = self._callback(data)
        if response is not None:
          log_line += ', responding with %s bytes' % len(response)
          # Send replies out-of-band instead of with the same multicast socket
          # so that multiple processes on the same host can listen for
          # requests and reply (if they all try to use the multicast socket
          # to reply, they conflict and this sendto fails).
          response = response.encode('utf-8')
          socket.socket(socket.AF_INET, socket.SOCK_DGRAM).sendto(
              response, address)
        _LOG.debug(log_line)
      except socket.timeout:
        pass
      except socket.error:
        _LOG.debug('Error receiving multicast message', exc_info=True)


class MulticastSendSocket(object):
  def __init__(self):
    self._sock = None
    
  def send(self, *args, **kwargs):
    if self._sock is None:
      self._sock = init_send_sock()
      
    return send(*args, _sock=self._sock, **kwargs)
  
def init_send_sock():
  sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
  return sock

def send(query,
         address=DEFAULT_ADDRESS,
         port=DEFAULT_PORT,
         ttl=DEFAULT_TTL,
         local_only=False,
         timeout_s=2,
         _sock=None):
  """Sends a query to the given multicast socket and returns responses.

  Args:
    query: The string query to send.
    address: Multicast IP address component of the socket to send to.
    port: Multicast UDP port component of the socket to send to.
    ttl: TTL for multicast messages. 1 to keep traffic in-network.
    timeout_s: Seconds to wait for responses.

  Returns: A set of all responses that arrived before the timeout expired.
           Responses are tuples of (sender_address, message).
  """
  # Set up the socket as a UDP Multicast socket with the given timeout.
  if _sock is None:
    _sock = init_send_sock()
  
  sock=_sock
  sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, ttl)
  
  bind_addr = address
  
  # Set outgoing interface to first in list of all_interfaces OR localhost
  if local_only:
    interface_ip = LOCALHOST_ADDRESS
  else:
    interfaces = get_all_interfaces()
    if interfaces:
      # at least one
      interface_ip = interfaces[0]
    else:
      # None ? use localhost then...
      interface_ip = LOCALHOST_ADDRESS
    
  sock.setsockopt(
      socket.IPPROTO_IP,
      socket.IP_MULTICAST_IF,
      struct.pack('!L', interface_ip))
  
  sock.settimeout(timeout_s)
  
  sock.sendto(query.encode('utf-8'), (bind_addr, port))

  # Set up our thread-safe Queue for handling responses.
  recv_queue = queue.Queue()
  def _handle_responses():
    while True:
      try:
        data, address = sock.recvfrom(MAX_MESSAGE_BYTES)
        data = data.decode('utf-8')
      except socket.timeout:
        recv_queue.put(None)
        break
      else:
        _LOG.debug('Multicast response to query "%s": %s:%s',
                   query, address[0], data)
        recv_queue.put((address[0], str(data)))

  # Yield responses as they come in, giving up once timeout expires.
  response_thread = threading.Thread(target=_handle_responses)
  response_thread.start()
  while response_thread.is_alive():
    recv_tuple = recv_queue.get()
    if not recv_tuple:
      break
    yield recv_tuple
  response_thread.join()



"""Convert a TestRecord into a mfg_event proto for upload to mfg inspector.

Also includes utilities to handle multi-dim conversion into an attachment
and the reverse.

A decision had to be made on how to handle phases, measurements and attachments
with non-unique names.  Approach taken is to append a _X to the names.
"""

import collections
import itertools
import json
import logging
import numbers
import os
import sys

from openhtf.core import measurements
from openhtf.core import test_record as htf_test_record
from openhtf.output.proto import mfg_event_pb2
from openhtf.output.proto import test_runs_converter
from openhtf.output.proto import test_runs_pb2
from openhtf.util import data as htf_data
from openhtf.util import units
from openhtf.util import validators


from past.builtins import unicode


TEST_RECORD_ATTACHMENT_NAME = 'OpenHTF_record.json'

#  To be lazy loaded by _LazyLoadUnitsByCode when needed.
UNITS_BY_CODE = {}

# Map test run Status (proto) name to measurement Outcome (python) enum's and
# the reverse.  Note: there is data lost in converting an UNSET/PARTIALLY_SET to
# an ERROR so we can't completely reverse the transformation.
MEASUREMENT_OUTCOME_TO_TEST_RUN_STATUS_NAME = {
    measurements.Outcome.PASS: 'PASS',
    measurements.Outcome.FAIL: 'FAIL',
    measurements.Outcome.UNSET: 'ERROR',
    measurements.Outcome.PARTIALLY_SET: 'ERROR',
}
TEST_RUN_STATUS_NAME_TO_MEASUREMENT_OUTCOME = {
    'PASS': measurements.Outcome.PASS,
    'FAIL': measurements.Outcome.FAIL,
    'ERROR': measurements.Outcome.UNSET
}


def _lazy_load_units_by_code():
  """Populate dict of units by code iff UNITS_BY_CODE is empty."""
  if UNITS_BY_CODE:
    # already populated
    return

  for unit in units.UNITS_BY_NAME.values():
    UNITS_BY_CODE[unit.code] = unit


def mfg_event_from_test_record(record):
  """Convert an OpenHTF TestRecord to an MfgEvent proto.

  Most fields are copied over directly and some are pulled out of metadata
  (listed below). Multi-dimensional measurements are stored only in the JSON
  dump of the record.

  Important Note:  This function mutates the test_record so any output callbacks
  called after this callback will operate on the mutated record.

  Metadata fields:
    test_name: The name field from the test's TestOptions.
    config: The OpenHTF config, as a dictionary.
    assembly_events: List of AssemblyEvent protos.
        (see proto/assembly_event.proto).
    operator_name: Name of the test operator.

  Args:
    record: An OpenHTF TestRecord.

  Returns:
    An MfgEvent proto representing the given test record.
  """
  mfg_event = mfg_event_pb2.MfgEvent()

  _populate_basic_data(mfg_event, record)
  _attach_record_as_json(mfg_event, record)
  _attach_argv(mfg_event)
  _attach_config(mfg_event, record)

  # Only include assembly events if the test passed.
  if ('assembly_events' in record.metadata and
      mfg_event.test_status == test_runs_pb2.PASS):
    for assembly_event in record.metadata['assembly_events']:
      mfg_event.assembly_events.add().CopyFrom(assembly_event)
  convert_multidim_measurements(record.phases)
  phase_copier = PhaseCopier(phase_uniquizer(record.phases))
  phase_copier.copy_measurements(mfg_event)
  phase_copier.copy_attachments(mfg_event)

  return mfg_event


def _populate_basic_data(mfg_event, record):
  """Copies data from the OpenHTF TestRecord to the MfgEvent proto."""
  # TODO:
  #   * Missing in proto: set run name from metadata.
  #   * `part_tags` field on proto is unused
  #   * `timings` field on proto is unused.
  #   * Handle arbitrary units as uom_code/uom_suffix.

  # Populate non-repeated fields.
  mfg_event.dut_serial = record.dut_id
  mfg_event.start_time_ms = record.start_time_millis
  mfg_event.end_time_ms = record.end_time_millis
  mfg_event.tester_name = record.station_id
  mfg_event.test_name = record.metadata.get('test_name') or record.station_id
  mfg_event.test_status = test_runs_converter.OUTCOME_MAP[record.outcome]
  mfg_event.operator_name = record.metadata.get('operator_name', '')
  mfg_event.test_version = str(record.metadata.get('test_version', ''))
  mfg_event.test_description = record.metadata.get('test_description', '')

  # Populate part_tags.
  mfg_event.part_tags.extend(record.metadata.get('part_tags', []))

  # Populate phases.
  for phase in record.phases:
    mfg_phase = mfg_event.phases.add()
    mfg_phase.name = phase.name
    mfg_phase.description = phase.codeinfo.sourcecode
    mfg_phase.timing.start_time_millis = phase.start_time_millis
    mfg_phase.timing.end_time_millis = phase.end_time_millis

  # Populate failure codes.
  for details in record.outcome_details:
    failure_code = mfg_event.failure_codes.add()
    failure_code.code = details.code
    failure_code.details = details.description

  # Populate test logs.
  for log_record in record.log_records:
    test_log = mfg_event.test_logs.add()
    test_log.timestamp_millis = log_record.timestamp_millis
    test_log.log_message = log_record.message
    test_log.logger_name = log_record.logger_name
    test_log.levelno = log_record.level
    if log_record.level <= logging.DEBUG:
      test_log.level = test_runs_pb2.TestRunLogMessage.DEBUG
    elif log_record.level <= logging.INFO:
      test_log.level = test_runs_pb2.TestRunLogMessage.INFO
    elif log_record.level <= logging.WARNING:
      test_log.level = test_runs_pb2.TestRunLogMessage.WARNING
    elif log_record.level <= logging.ERROR:
      test_log.level = test_runs_pb2.TestRunLogMessage.ERROR
    elif log_record.level <= logging.CRITICAL:
      test_log.level = test_runs_pb2.TestRunLogMessage.CRITICAL
    test_log.log_source = log_record.source
    test_log.lineno = log_record.lineno


def _attach_record_as_json(mfg_event, record):
  """Attach a copy of the record as JSON so we have an un-mangled copy."""
  attachment = mfg_event.attachment.add()
  attachment.name = TEST_RECORD_ATTACHMENT_NAME
  test_record_dict = htf_data.convert_to_base_types(record)
  attachment.value_binary = _convert_object_to_json(test_record_dict)
  attachment.type = test_runs_pb2.TEXT_UTF8


def _convert_object_to_json(obj):
  # Since there will be parts of this that may have unicode, either as
  # measurement or in the logs, we have to be careful and convert everything
  # to unicode, merge, then encode to UTF-8 to put it into the proto.
  json_encoder = json.JSONEncoder(sort_keys=True, indent=2, ensure_ascii=False)
  pieces = []
  for piece in json_encoder.iterencode(obj):
    if isinstance(piece, bytes):
      pieces.append(unicode(piece, errors='replace'))
    else:
      pieces.append(piece)

  return (u''.join(pieces)).encode('utf8', errors='replace')


def _attach_config(mfg_event, record):
  """Attaches the OpenHTF config file as JSON."""
  if 'config' not in record.metadata:
    return
  attachment = mfg_event.attachment.add()
  attachment.name = 'config'
  attachment.value_binary = _convert_object_to_json(record.metadata['config'])
  attachment.type = test_runs_pb2.TEXT_UTF8


def _attach_argv(mfg_event):
  attachment = mfg_event.attachment.add()
  attachment.name = 'argv'
  argv = [os.path.realpath(sys.argv[0])] + sys.argv[1:]
  attachment.value_binary = _convert_object_to_json(argv)
  attachment.type = test_runs_pb2.TEXT_UTF8


class UniqueNameMaker(object):
  """Makes unique names for phases, attachments, etc with duplicate names."""

  def __init__(self, all_names):
    self._counts = collections.Counter(all_names)
    self._seen = collections.Counter()

  def make_unique(self, name):
    count = self._counts[name]
    assert count >= 1, 'Seeing a new name that was not given to the constructor'
    if count == 1:
      # It's unique, so let's skip extra calculations.
      return name
    # Count the number of times we've seen this and return this one's index.
    self._seen[name] += 1
    main, ext = os.path.splitext(name)

    return '%s_%d%s' % (main, self._seen[name] - 1, ext)


def phase_uniquizer(all_phases):
  """Makes the names of phase measurement and attachments unique.

  This function will make the names of measurements and attachments unique.
  It modifies the input all_phases.

  Args:
    all_phases: the phases to make unique

  Returns:
    the phases now modified.
  """
  measurement_name_maker = UniqueNameMaker(
      itertools.chain.from_iterable(
          phase.measurements.keys() for phase in all_phases
          if phase.measurements))
  attachment_names = list(itertools.chain.from_iterable(
      phase.attachments.keys() for phase in all_phases))
  attachment_names.extend(itertools.chain.from_iterable([
      'multidim_' + name for name, meas in phase.measurements.items()
      if meas.dimensions is not None
  ] for phase in all_phases if phase.measurements))
  attachment_name_maker = UniqueNameMaker(attachment_names)
  for phase in all_phases:
    # Make measurements unique.
    for name, _ in sorted(phase.measurements.items()):
      old_name = name
      name = measurement_name_maker.make_unique(name)

      phase.measurements[old_name].name = name
      phase.measurements[name] = phase.measurements.pop(old_name)
    # Make attachments unique.
    for name, _ in sorted(phase.attachments.items()):
      old_name = name
      name = attachment_name_maker.make_unique(name)
      phase.attachments[old_name].name = name
      phase.attachments[name] = phase.attachments.pop(old_name)
  return all_phases


def multidim_measurement_to_attachment(name, measurement):
  """Convert a multi-dim measurement to an `openhtf.test_record.Attachment`."""

  dimensions = list(measurement.dimensions)
  if measurement.units:
    dimensions.append(
        measurements.Dimension.from_unit_descriptor(measurement.units))

  dims = []
  for d in dimensions:
    if d.suffix is None:
      suffix = u''
    # Ensure that the suffix is unicode. It's typically str/bytes because
    # units.py looks them up against str/bytes.
    elif isinstance(d.suffix, unicode):
      suffix = d.suffix
    else:
      suffix = d.suffix.decode('utf8')
    dims.append({
        'uom_suffix': suffix,
        'uom_code': d.code,
        'name': d.name,
    })
  # Refer to the module docstring for the expected schema.
  dimensioned_measured_value = measurement.measured_value
  value = (sorted(dimensioned_measured_value.value, key=lambda x: x[0])
           if dimensioned_measured_value.is_value_set else None)
  outcome_str = MEASUREMENT_OUTCOME_TO_TEST_RUN_STATUS_NAME[measurement.outcome]
  data = _convert_object_to_json({
      'outcome': outcome_str,
      'name': name,
      'dimensions': dims,
      'value': value,
  })
  attachment = htf_test_record.Attachment(data, test_runs_pb2.MULTIDIM_JSON)

  return attachment


def convert_multidim_measurements(all_phases):
  """Converts each multidim measurements into attachments for all phases.."""
  # Combine actual attachments with attachments we make from multi-dim
  # measurements.
  attachment_names = list(itertools.chain.from_iterable(
      phase.attachments.keys() for phase in all_phases))
  attachment_names.extend(itertools.chain.from_iterable([
      'multidim_' + name for name, meas in phase.measurements.items()
      if meas.dimensions is not None
  ] for phase in all_phases if phase.measurements))
  attachment_name_maker = UniqueNameMaker(attachment_names)

  for phase in all_phases:
    # Process multi-dim measurements into unique attachments.
    for name, measurement in sorted(phase.measurements.items()):
      if measurement.dimensions:
        old_name = name
        name = attachment_name_maker.make_unique('multidim_%s' % name)
        attachment = multidim_measurement_to_attachment(name, measurement)
        phase.attachments[name] = attachment
        phase.measurements.pop(old_name)
  return all_phases


class PhaseCopier(object):
  """Copies measurements and attachments to an MfgEvent."""

  def __init__(self, all_phases):
    self._phases = all_phases

  def copy_measurements(self, mfg_event):
    for phase in self._phases:
      for name, measurement in sorted(phase.measurements.items()):
        # Multi-dim measurements should already have been removed.
        assert measurement.dimensions is None
        self._copy_unidimensional_measurement(
            phase, name, measurement, mfg_event)

  def _copy_unidimensional_measurement(
      self, phase, name, measurement, mfg_event):
    """Copy uni-dimensional measurements to the MfgEvent."""
    mfg_measurement = mfg_event.measurement.add()

    # Copy basic measurement fields.
    mfg_measurement.name = name
    if measurement.docstring:
      mfg_measurement.description = measurement.docstring
    mfg_measurement.parameter_tag.append(phase.name)
    if (measurement.units and
        measurement.units.code in test_runs_converter.UOM_CODE_MAP):
      mfg_measurement.unit_code = (
          test_runs_converter.UOM_CODE_MAP[measurement.units.code])

    # Copy failed measurements as failure_codes. This happens early to include
    # unset measurements.
    if (measurement.outcome != measurements.Outcome.PASS and
        phase.outcome != htf_test_record.PhaseOutcome.SKIP):
      failure_code = mfg_event.failure_codes.add()
      failure_code.code = name
      failure_code.details = '\n'.join(str(v) for v in measurement.validators)

    # Copy measurement value.
    measured_value = measurement.measured_value
    status_str = MEASUREMENT_OUTCOME_TO_TEST_RUN_STATUS_NAME[
        measurement.outcome]
    mfg_measurement.status = test_runs_pb2.Status.Value(status_str)
    if not measured_value.is_value_set:
      return
    value = measured_value.value

    if isinstance(value, numbers.Number):
      mfg_measurement.numeric_value = float(value)
    elif isinstance(value, bytes):
      # text_value expects unicode or ascii-compatible strings, so we must
      # 'decode' it, even if it's actually just garbage bytestring data.
      mfg_measurement.text_value = unicode(value, errors='replace')
    elif isinstance(value, unicode):
      # Don't waste time and potential errors decoding unicode.
      mfg_measurement.text_value = value
    else:
      # Coercing to string.
      mfg_measurement.text_value = str(value)

    # Copy measurement validators.
    for validator in measurement.validators:
      if isinstance(validator, validators.RangeValidatorBase):
        if validator.minimum is not None:
          mfg_measurement.numeric_minimum = float(validator.minimum)
        if validator.maximum is not None:
          mfg_measurement.numeric_maximum = float(validator.maximum)
      elif isinstance(validator, validators.RegexMatcher):
        mfg_measurement.expected_text = validator.regex
      else:
        mfg_measurement.description += '\nValidator: ' + str(validator)

  def copy_attachments(self, mfg_event):
    for phase in self._phases:
      for name, (data, mimetype) in sorted(phase.attachments.items()):
        self._copy_attachment(name, data, mimetype, mfg_event)

  def _copy_attachment(self, name, data, mimetype, mfg_event):
    """Copies an attachment to mfg_event."""
    attachment = mfg_event.attachment.add()
    attachment.name = name
    if isinstance(data, unicode):
      data = data.encode('utf8')
    attachment.value_binary = data
    if mimetype in test_runs_converter.MIMETYPE_MAP:
      attachment.type = test_runs_converter.MIMETYPE_MAP[mimetype]
    elif mimetype == test_runs_pb2.MULTIDIM_JSON:
      attachment.type = mimetype
    else:
      attachment.type = test_runs_pb2.BINARY


def test_record_from_mfg_event(mfg_event):
  """Extract the original test_record saved as an attachment on a mfg_event."""
  for attachment in mfg_event.attachment:
    if attachment.name == TEST_RECORD_ATTACHMENT_NAME:
      return json.loads(attachment.value_binary)

  raise ValueError('Could not find test record JSON in the given MfgEvent.')


def attachment_to_multidim_measurement(attachment, name=None):
  """Convert an OpenHTF test record attachment to a multi-dim measurement.

  This is a best effort attempt to reverse, as some data is lost in converting
  from a multidim to an attachment.

  Args:
    attachment: an `openhtf.test_record.Attachment` from a multi-dim.
    name: an optional name for the measurement.  If not provided will use the
     name included in the attachment.

  Returns:
    An multi-dim `openhtf.Measurement`.
  """
  data = json.loads(attachment.data)

  name = name or data.get('name')
  # attachment_dimn are a list of dicts with keys 'uom_suffix' and 'uom_code'
  attachment_dims = data.get('dimensions', [])
  # attachment_value is a list of lists [[t1, x1, y1, f1], [t2, x2, y2, f2]]
  attachment_values = data.get('value')

  attachment_outcome_str = data.get('outcome')
  if attachment_outcome_str not in TEST_RUN_STATUS_NAME_TO_MEASUREMENT_OUTCOME:
    # Fpr backward compatibility with saved data we'll convert integers to str
    try:
      attachment_outcome_str = test_runs_pb2.Status.Name(
          int(attachment_outcome_str))
    except ValueError:
      attachment_outcome_str = None

  # Convert test status outcome str to measurement outcome
  outcome = TEST_RUN_STATUS_NAME_TO_MEASUREMENT_OUTCOME.get(
      attachment_outcome_str)

  # convert dimensions into htf.Dimensions
  _lazy_load_units_by_code()
  dims = []
  for d in attachment_dims:
    # Try to convert into htf.Dimension including backwards compatibility.
    unit = UNITS_BY_CODE.get(d.get('uom_code'), units.NONE)
    description = d.get('name', '')
    dims.append(measurements.Dimension(description=description, unit=unit))

  # Attempt to determine if units are included.
  if attachment_values and len(dims) == len(attachment_values[0]):
    # units provided
    units_ = dims[-1].unit
    dimensions = dims[:-1]
  else:
    units_ = None
    dimensions = dims

  # created dimensioned_measured_value and populate with values.
  measured_value = measurements.DimensionedMeasuredValue(
      name=name,
      num_dimensions=len(dimensions)
  )
  for row in attachment_values:
    coordinates = tuple(row[:-1])
    val = row[-1]
    measured_value[coordinates] = val

  measurement = measurements.Measurement(
      name=name,
      units=units_,
      dimensions=tuple(dimensions),
      measured_value=measured_value,
      outcome=outcome
  )
  return measurement



# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""UsbHandle implementation using libusb to communicate with local devices.

This implementation of UsbHandle uses python libusb1 bindings to communicate
with locally attached USB devices.  In order to fit the UsbHandle interface,
much of the complexity of the libusb1 interface is abstracted away.  This class
should never raise libusb1 errors directly - if it does it is a bug, please fix
it and wrap the error in an appropriate exception from usb_exceptions.py.
"""

import logging

try:
  import libusb1
  import usb1
except ImportError:
  logging.error('Failed to import libusb, did you pip install '
                'openhtf[usb_plugs]?')
  raise

from openhtf.plugs.usb import usb_exceptions
from openhtf.plugs.usb import usb_handle
import six

_LOG = logging.getLogger(__name__)


class LibUsbHandle(usb_handle.UsbHandle):
  """Subclass of UsbHandle that opens locally connected devices with libusb."""

  def __init__(self, device, setting, name=None, default_timeout_ms=None):
    """Initialize a local libusb-based USB Handle.

    Arguments:
      device: libusb device to connect to.
      setting: libusb setting with the endpoints to use for data reads/writes.
      name: Name for the device, used for log messages only.
      default_timeout_ms: Default timeout, in milliseconds, for reads/writes.

    Raises:
      InvalidEndpointsError: If the setting provided does not have exactly one
        in endpoint and one out endpoint.  Currently, this is the only type of
        interface supported.
      IOError: If the device has been disconnected.
    """
    super(LibUsbHandle, self).__init__(device.getSerialNumber(), name=name,
                                       default_timeout_ms=default_timeout_ms)
    self._setting = setting
    self._device = device

    self._read_endpoint = None
    self._write_endpoint = None
    self._handle = None

    for endpoint in self._setting.iterEndpoints():
      address = endpoint.getAddress()
      if address & libusb1.USB_ENDPOINT_DIR_MASK == libusb1.LIBUSB_ENDPOINT_IN:
        if self._read_endpoint is not None:
          raise usb_exceptions.InvalidEndpointsError(
              'Multiple in endpoints found')
        self._read_endpoint = address
      else:
        if self._write_endpoint is not None:
          raise usb_exceptions.InvalidEndpointsError(
              'Multiple out endpoints found')
        self._write_endpoint = address

    if self._read_endpoint is None:
      raise usb_exceptions.InvalidEndpointsError('Missing in endpoint')
    if self._write_endpoint is None:
      raise usb_exceptions.InvalidEndpointsError('Missing out endpoint')

    handle = self._device.open()
    iface_number = self._setting.getNumber()
    try:
      if handle.kernelDriverActive(iface_number):
        handle.detachKernelDriver(iface_number)
    except libusb1.USBError as exception:
      if exception.value == libusb1.LIBUSB_ERROR_NOT_FOUND:
        _LOG.warning('Kernel driver not found for interface: %s.', iface_number)
      else:
        raise usb_exceptions.LibusbWrappingError(
            exception, '%s failed to detach kernel driver', self)
    # Try to claim the interface, if the device is gone, raise an IOError
    try:
      handle.claimInterface(iface_number)
    except libusb1.USBError as exception:
      raise usb_exceptions.LibusbWrappingError(
          exception, '%s failed to claim interface: %d', self, iface_number)

    self._handle = handle
    self._interface_number = iface_number

  def is_closed(self):
    return self._handle is None

  @staticmethod
  def _device_to_sysfs_path(device):
    """Convert device to corresponding sysfs path."""
    return '%s-%s' % (
        device.getBusNumber(),
        '.'.join([str(item) for item in device.GetPortNumberList()]))

  @property
  def port_path(self):
    """A string of the physical port of this device, like 'X-X.X.X'."""
    return self._device_to_sysfs_path(self._device)

  @usb_handle.requires_open_handle
  def read(self, length, timeout_ms=None):
    try:
      return self._handle.bulkRead(
          self._read_endpoint, length,
          timeout=self._timeout_or_default(timeout_ms))
    except libusb1.USBError as exception:
      raise usb_exceptions.UsbReadFailedError(
          exception, '%s failed read (timeout %sms)', self,
          self._timeout_or_default(timeout_ms))

  @usb_handle.requires_open_handle
  def write(self, data, timeout_ms=None):
    try:
      return self._handle.bulkWrite(
          self._write_endpoint, data,
          timeout=self._timeout_or_default(timeout_ms))
    except libusb1.USBError as exception:
      raise usb_exceptions.UsbWriteFailedError(
          exception, '%s failed write (timeout %sms)', self,
          self._timeout_or_default(timeout_ms))

  def close(self):
    if self.is_closed():
      return

    try:
      self._handle.releaseInterface(self._interface_number)
      self._handle.close()
    except libusb1.USBError:
      _LOG.exception('USBError while closing handle %s:', self)
    finally:
      self._handle = None

  @classmethod
  def open(cls, **kwargs):
    """See iter_open, but raises if multiple or no matches found."""
    handle_iter = cls.iter_open(**kwargs)

    try:
      handle = six.next(handle_iter)
    except StopIteration:
      # No matching interface, raise.
      raise usb_exceptions.DeviceNotFoundError(
          'Open failed with args: %s', kwargs)

    try:
      multiple_handle = six.next(handle_iter)
    except StopIteration:
      # Exactly one matching device, return it.
      return handle

    # We have more than one device, close the ones we opened and bail.
    handle.close()
    multiple_handle.close()
    raise usb_exceptions.MultipleInterfacesFoundError(kwargs)

  # pylint: disable=too-many-arguments
  @classmethod
  def iter_open(cls, name=None, interface_class=None, interface_subclass=None,
                interface_protocol=None, serial_number=None, port_path=None,
                default_timeout_ms=None):
    """Find and yield locally connected devices that match.

    Note that devices are opened (and interfaces claimd) as they are yielded.
    Any devices yielded must be Close()'d.

    Args:
      name: Name to give *all* returned handles, used for logging only.
      interface_class: USB interface_class to match.
      interface_subclass: USB interface_subclass to match.
      interface_protocol: USB interface_protocol to match.
      serial_number: USB serial_number to match.
      port_path: USB Port path to match, like X-X.X.X
      default_timeout_ms: Default timeout in milliseconds of reads/writes on
        the handles yielded.

    Yields:
      UsbHandle instances that match any non-None args given.

    Raises:
      LibusbWrappingError: When a libusb call errors during open.
    """
    ctx = usb1.USBContext()
    try:
      devices = ctx.getDeviceList(skip_on_error=True)
    except libusb1.USBError as exception:
      raise usb_exceptions.LibusbWrappingError(
          exception, 'Open(name=%s, class=%s, subclass=%s, protocol=%s, '
          'serial=%s, port=%s) failed', name, interface_class,
          interface_subclass, interface_protocol, serial_number, port_path)

    for device in devices:
      try:
        if (serial_number is not None and
            device.getSerialNumber() != serial_number):
          continue

        if (port_path is not None and
            cls._device_to_sysfs_path(device) != port_path):
          continue

        for setting in device.iterSettings():
          if (interface_class is not None and
              setting.getClass() != interface_class):
            continue
          if (interface_subclass is not None and
              setting.getSubClass() != interface_subclass):
            continue
          if (interface_protocol is not None and
              setting.getProtocol() != interface_protocol):
            continue

          yield cls(device, setting, name=name,
                    default_timeout_ms=default_timeout_ms)
      except libusb1.USBError as exception:
        if (exception.value !=
            libusb1.libusb_error.forward_dict['LIBUSB_ERROR_ACCESS']):
          raise
  # pylint: disable=too-many-arguments



import logging
import gspread
import datetime

from oauth2client.service_account import ServiceAccountCredentials
from googleapiclient.discovery import build

_LOG = logging.getLogger(__name__)

class FileNotFound(Exception): """ The Google drive file was not found. """

def load_credentials_file(credentials_file, scope):
    return ServiceAccountCredentials.from_json_keyfile_name(credentials_file, scope)

class GoogleDrive(object):
    def __init__(self, credentials):
        self.credentials = credentials
        self._service = None
        self._gc = None

    @property
    def service(self):
        """ Lazy loaded. This allows network to not yet be available when created. """
        if self._service is None:
            self._service = build('drive', 'v3', credentials=self.credentials)
        return self._service

    @property
    def gc(self):
        """ Lazy loaded. This allows network to not yet be available when created. """
        if self._gc is None:
            self._gc = gspread.authorize(self.credentials)
        return self._gc



    def print_files(self):
        results = self.service.files().list(
            pageSize=10, fields="nextPageToken, files(id, name, webViewLink)").execute()
        items = results.get('files', [])

        if not items:
            print('No files found.')
        else:
            print('Files:')
            for item in items:
                print(u'{0} ({1})'.format(item['name'], item['id']))

    def get_or_create_folder(self, name, share_with=[]):
        folder = None
        try:
            folder_id = self.find_folder(name)
        except FileNotFound:
            folder_id = self.create_folder(name)
        permissions = self.service.permissions().list(fileId=folder_id, fields="permissions(emailAddress)").execute()['permissions']
        current_email_addresses = [perm['emailAddress'] for perm in permissions]
        for email in share_with:
            if email not in current_email_addresses:
                self.share_folder(folder_id, email)
        return folder_id


    def find_folder(self, folder_name):
        response = self.service.files().list(q="mimeType='application/vnd.google-apps.folder' and name='%s'" % folder_name,
                                          spaces='drive',
                                          fields='files(id, name)').execute()
        for f in response.get('files', []):
            return f['id']
        else:
            raise FileNotFound(folder_name)

    def create_folder(self, folder_name, parent_id = None):
        # Create a folder on Drive, returns the newely created folders ID
        body = {
          'name': folder_name,
          'mimeType': "application/vnd.google-apps.folder"
        }
        if parent_id:
            body['parents'] = [parent_id]
        root_folder = self.service.files().create(body = body).execute()
        return root_folder['id']

    def move_file_to_folder(self, file_id, folder_id):
        # Retrieve the existing parents to remove
        file = self.service.files().get(fileId=file_id,
                                        fields='parents').execute()
        previous_parents = ",".join(file.get('parents'))
        # Move the file to the new folder
        file = self.service.files().update(fileId=file_id,
                                            addParents=folder_id,
                                            removeParents=previous_parents,
                                            fields='id, parents').execute()

    def find_filename_in_folder(self, filename, folder_id):
        q = "name='%s'" % filename
        if folder_id:
            q = ("'%s' in parents and " % folder_id) + q

        response = self.service.files().list(q=q,
                                          spaces='drive',
                                          fields='files(id, name)').execute()
        for f in response.get('files', []):
            return f['id']
        else:
            raise FileNotFound('%s in %s' % (filename, folder_id))


    def share_folder(self, folder_id, email):
        def callback(request_id, response, exception):
            if exception:
                # Handle error
                print(exception)
            else:
                print("Permission Id: %s" % response.get('id'))

        batch = self.service.new_batch_http_request(callback=callback)
        user_permission = {
            'type': 'user',
            'role': 'writer',
            'emailAddress': email
        }
        batch.add(self.service.permissions().create(
                fileId=folder_id,
                body=user_permission,
                fields='id',
        ))
        batch.execute()

    def get_or_create_spreadsheet(self, filename, folder_id=None):
        try:
            file_id = self.find_filename_in_folder(filename, folder_id)
            spreadsheet = self.gc.open_by_key(file_id)
        except FileNotFound:
            spreadsheet = self.gc.create(filename)
            if folder_id:
                self.move_file_to_folder(spreadsheet.id, folder_id)
        
        return spreadsheet

class GoogleFolder(object):
    def __init__(self, gdrive, folder_id):
        self.gdrive = gdrive
        self.folder_id = folder_id

    def get_or_create_spreadsheet(self, filename):
        return self.gdrive.get_or_create_spreadsheet(filename, folder_id=self.folder_id)
    

class Worksheet(object):
    def __init__(self, spreadsheet, title, headers=[], first=False):
        self.spreadsheet = spreadsheet
        self.sheet = None
        self.headers = headers

        self._open_worksheet(title, first=first)

    def _open_worksheet(self, title, first=False):
        if first:
            self.sheet = self.spreadsheet.get_worksheet(0)
            self.sheet.update_title(title)
        else:
            try:
                self.sheet = self.spreadsheet.worksheet(title)
            except gspread.WorksheetNotFound:
                self.sheet = self.spreadsheet.add_worksheet(title=title, rows="100", cols="100")

        self.headers = self.update_headers(self.headers)

    def update_headers(self, headers):
        current_headers = self.sheet.row_values(1)
        if not headers:
            headers = []

        if current_headers:
            headers = current_headers + [h for h in headers if h not in current_headers]
            self.sheet.delete_row(1)

        if headers:
            self.sheet.insert_row(headers)

        return headers

    def append_row(self, row_d, value_input_option='USER_ENTERED'):
        row = self._row_d_to_row(row_d)

        self.sheet.append_row(row, value_input_option=value_input_option)

    def _row_d_to_row(self, row_d):
        """ Transform a dict row to a list in order based on current headers (self.headers)
        headers = ['a', 'b', 'c']
        row_d = {'a': 1, 'c': 2}

        Returns => [1, "", 2]
        """
        return [row_d.get(header, "") for header in self.headers]

    def append_rows(self, rows, value_input_option='USER_ENTERED'):
        params = {
            'valueInputOption': value_input_option
        }

        body = {
            'values': [self._row_d_to_row(row_d) for row_d in rows]
        }

        return self.sheet.spreadsheet.values_append(self.sheet.title, params, body)

